from numba import float64
from numba import int64
from numba import njit

import numpy as np
from numpy.typing import NDArray


@njit(float64(float64[:], float64[:]))
def root_mean_square_error(
    y_true: NDArray[np.float64], y_predict: NDArray[np.float64]
) -> np.float64:
    error = y_true - y_predict
    return np.sqrt(np.mean((error) ** 2))


@njit(float64[:](float64[:], float64[:, :]))
def root_mean_square_error2d(
    y_true: NDArray[np.float64], y_predict2d: NDArray[np.float64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    to_return = np.empty(size, dtype=np.float64)

    for i in range(size):
        error = y_true - y_predict2d[i]
        to_return[i] = np.sqrt(np.mean((error) ** 2))
    return to_return


@njit(float64(float64[:], float64[:]))
def coefficient_determination(
    y_true: NDArray[np.float64], y_predict: NDArray[np.float64]
) -> np.float64:
    error = y_true - y_predict
    mean_y_true = np.mean(y_true)
    residual_sum = np.sum((error) ** 2)
    total_sum = np.sum((y_true - mean_y_true) ** 2)
    return 1 - residual_sum / total_sum


@njit(float64[:](float64[:], float64[:, :]))
def coefficient_determination2d(
    y_true: NDArray[np.float64], y_predict2d: NDArray[np.float64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    to_return = np.empty(size, dtype=np.float64)

    mean_y_true = np.mean(y_true)
    total_sum = np.sum((y_true - mean_y_true) ** 2)
    for i in range(size):
        error = y_true - y_predict2d[i]
        residual_sum = np.sum((error) ** 2)
        to_return[i] = 1 - residual_sum / total_sum
    return to_return


@njit(float64(float64[:, :], float64[:, :]))
def categorical_crossentropy(
    target: NDArray[np.float64], output: NDArray[np.float64]
) -> np.float64:
    output_c = np.clip(output, 1e-7, 1 - 1e-7)
    to_return = np.mean(np.sum(target * (-np.log(output_c)), axis=1))
    return to_return


@njit(float64[:](float64[:, :], float64[:, :, :]))
def categorical_crossentropy3d(
    target: NDArray[np.float64], output3d: NDArray[np.float64]
) -> NDArray[np.float64]:
    size = len(output3d)
    to_return = np.empty(size, dtype=np.float64)
    for i in range(size):
        to_return[i] = categorical_crossentropy(target, output3d[i])
    return to_return


@njit(float64(int64[:], int64[:]))
def accuracy_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    return np.mean((y_true == y_predict).astype(np.int64))


@njit(float64[:](int64[:], int64[:, :]))
def accuracy_score2d(
    y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    to_return = np.empty(size, dtype=np.float64)
    for i in range(size):
        to_return[i] = accuracy_score(y_true, y_predict2d[i])
    return to_return


@njit(int64[:, :](int64[:], int64[:]))
def confusion_matrix(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> NDArray[np.int64]:
    """
    y_true = np.array([0, 1, 0, 2 ...], dtype = np.int64)
    y_predict = np.array([1, 2, 0, 2 ...], dtype = np.int64)
    """
    classes = np.unique(y_true)
    n_classes = len(classes)
    size = len(y_true)

    to_return = np.zeros(shape=(n_classes, n_classes), dtype=np.int64)
    for i in range(size):
        to_return[y_true[i]][y_predict[i]] += 1
    return to_return


@njit(float64(int64[:], int64[:]))
def recall_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    """
    y_true = np.array([0, 1, 0, 2 ...], dtype = np.int64)
    y_predict = np.array([1, 2, 0, 2 ...], dtype = np.int64)
    """
    classes = np.unique(y_true)
    n_classes = len(classes)
    size = len(y_true)

    up = np.zeros(shape=n_classes, dtype=np.int64)
    down = np.zeros(shape=n_classes, dtype=np.int64)

    for i in range(size):
        if y_true[i] == y_predict[i]:
            up[y_true[i]] += 1
        else:
            down[y_true[i]] += 1

    recall = up / (down + up)
    recall[np.isnan(recall)] = 1

    return np.mean(recall)


@njit(float64[:](int64[:], int64[:, :]))
def recall_score2d(
    y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    to_return = np.empty(size, dtype=np.float64)
    for i in range(size):
        to_return[i] = recall_score(y_true, y_predict2d[i])
    return to_return


@njit(float64(int64[:], int64[:]))
def precision_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    """
    y_true = np.array([0, 1, 0, 2 ...], dtype = np.int64)
    y_predict = np.array([1, 2, 0, 2 ...], dtype = np.int64)
    """
    classes = np.unique(y_true)
    n_classes = len(classes)
    size = len(y_true)

    up = np.zeros(shape=n_classes, dtype=np.int64)
    down = np.zeros(shape=n_classes, dtype=np.int64)

    for i in range(size):
        if y_true[i] == y_predict[i]:
            up[y_true[i]] += 1
        else:
            down[y_predict[i]] += 1

    precision = up / (down + up)
    precision[np.isnan(precision)] = 1

    return np.mean(precision)


@njit(float64[:](int64[:], int64[:, :]))
def precision_score2d(
    y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]
) -> NDArray[np.float64]:
    size = len(y_predict2d)
    to_return = np.empty(size, dtype=np.float64)
    for i in range(size):
        to_return[i] = precision_score(y_true, y_predict2d[i])
    return to_return


@njit(float64(int64[:], int64[:]))
def f1_score(y_true: NDArray[np.int64], y_predict: NDArray[np.int64]) -> np.float64:
    """
    y_true = np.array([0, 1, 0, 2 ...], dtype = np.int64)
    y_predict = np.array([1, 2, 0, 2 ...], dtype = np.int64)
    """
    classes = np.unique(y_true)
    n_classes = len(classes)
    size = len(y_true)

    up = np.zeros(shape=n_classes, dtype=np.int64)
    down_recall = np.zeros(shape=n_classes, dtype=np.int64)
    down_precision = np.zeros(shape=n_classes, dtype=np.int64)

    for i in range(size):
        if y_true[i] == y_predict[i]:
            up[y_true[i]] += 1
        else:
            down_recall[y_true[i]] += 1
            down_precision[y_predict[i]] += 1

    precision = up / (down_precision + up)
    recall = up / (down_recall + up)

    precision[np.isnan(precision)] = 1
    recall[np.isnan(recall)] = 1

    f1 = 2 * (precision * recall) / (precision + recall)
    return np.mean(f1)


@njit(float64[:](int64[:], int64[:, :]))
def f1_score2d(y_true: NDArray[np.int64], y_predict2d: NDArray[np.int64]) -> NDArray[np.float64]:
    size = len(y_predict2d)
    to_return = np.empty(size, dtype=np.float64)
    for i in range(size):
        to_return[i] = f1_score(y_true, y_predict2d[i])
    return to_return



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from ._geneticalgorithm import GeneticAlgorithm
from ..base import Tree
from ..base import UniversalSet
from ..tools import donothing
from ..tools.random import half_and_half


class GeneticProgramming(GeneticAlgorithm):
    """Koza, John R.. “Genetic programming - on the programming of computers by means
    of natural selection.” Complex Adaptive Systems (1993)"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        uniset: UniversalSet,
        iters: int,
        pop_size: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 7,
        elitism: bool = True,
        selection: str = "rank",
        crossover: str = "gp_standart",
        mutation: str = "gp_weak_grow",
        max_level: int = 16,
        init_level: int = 5,
        init_population: Optional[NDArray] = None,
        genotype_to_phenotype: Callable[[NDArray[Any]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        GeneticAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=1,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            selection=selection,
            crossover=crossover,
            mutation=mutation,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        self._uniset: UniversalSet = uniset
        self._max_level: int = max_level
        self._init_level: int = init_level

    def _get_init_population(self: GeneticProgramming) -> None:
        if self._init_population is None:
            self._population_g_i = self._population_g_i = half_and_half(
                pop_size=self._pop_size, uniset=self._uniset, max_level=self._init_level
            )
        else:
            self._population_g_i = self._init_population.copy()

    def _get_new_individ_g(
        self: GeneticProgramming,
        specified_selection: str,
        specified_crossover: str,
        specified_mutation: str,
    ) -> Tree:
        selection_func, tour_size = self._selection_pool[specified_selection]
        crossover_func, quantity = self._crossover_pool[specified_crossover]
        mutation_func, proba, is_constant_rate = self._mutation_pool[specified_mutation]

        selected_id = selection_func(
            self._fitness_scale_i, self._fitness_rank_i, np.int64(tour_size), np.int64(quantity)
        )

        offspring_no_mutated = crossover_func(
            self._population_g_i[selected_id],
            self._fitness_scale_i[selected_id],
            self._fitness_rank_i[selected_id],
            self._max_level,
        )

        if is_constant_rate:
            proba = proba
        else:
            proba = proba / len(offspring_no_mutated)

        offspring = mutation_func(offspring_no_mutated, self._uniset, proba, self._max_level)
        return offspring



from typing import Any

from ._numba_funcs import argsort_k
from ._numba_funcs import binary_search_interval
from ._numba_funcs import check_for_value
from ._numba_funcs import find_end_subtree_from_i
from ._numba_funcs import find_first_difference_between_two
from ._numba_funcs import find_id_args_from_i
from ._numba_funcs import find_pbest_id
from ._numba_funcs import get_levels_tree_from_i


def donothing(x: Any) -> Any:
    return x


__all__ = [
    "find_end_subtree_from_i",
    "find_id_args_from_i",
    "get_levels_tree_from_i",
    "find_first_difference_between_two",
    "binary_search_interval",
    "check_for_value",
    "argsort_k",
    "find_pbest_id",
]



from __future__ import annotations

from typing import Any
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray


class Model:
    def _fit(
        self,
        X: NDArray[np.float64],
        y: NDArray[Union[np.float64, np.int64]],
    ) -> Any:
        pass

    def _predict(self, X: NDArray[np.float64]) -> Any:
        pass

    def get_optimizers(self: Model) -> Tuple:
        return tuple()

    def fit(
        self,
        X: NDArray[np.float64],
        y: NDArray[Union[np.float64, np.int64]],
    ) -> Any:
        return self._fit(X, y)

    def predict(self, X: NDArray[np.float64]) -> NDArray[Union[np.float64, np.int64]]:
        return self._predict(X)



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from ._differentialevolution import DifferentialEvolution
from ..tools import donothing


class jDE(DifferentialEvolution):
    """Brest, Janez & Greiner, Sao & Bošković, Borko & Mernik, Marjan & Zumer, Viljem. (2007).
    Self-Adapting Control Parameters in Differential Evolution: A Comparative Study on Numerical
    Benchmark Problems. Evolutionary Computation, IEEE Transactions on. 10.
      646 - 657. 10.1109/TEVC.2006.872133."""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        left: NDArray[np.float64],
        right: NDArray[np.float64],
        mutation: str = "rand_1",
        F_min: float = 0.1,
        F_max: float = 0.9,
        t_F: float = 0.1,
        t_CR: float = 0.1,
        elitism: bool = True,
        init_population: Optional[NDArray[np.float64]] = None,
        genotype_to_phenotype: Callable[[NDArray[np.float64]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            left=left,
            right=right,
            mutation=mutation,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        self._F_min: float = F_min
        self._F_max: float = F_max
        self._t_F: float = t_F
        self._t_CR: float = t_CR

        self._F: NDArray[np.float64] = np.full(self._pop_size, 0.5, dtype=np.float64)
        self._CR: NDArray[np.float64] = np.full(self._pop_size, 0.9, dtype=np.float64)

    def _get_mutate_F(self: jDE) -> NDArray[np.float64]:
        mutate_F = self._F.copy()
        mask = np.random.random(size=self._pop_size) < self._t_F

        random_values = np.random.random(size=np.sum(mask))
        mutate_F[mask] = self._F_min + random_values * self._F_max
        return mutate_F

    def _get_mutate_CR(self: jDE) -> NDArray[np.float64]:
        mutate_CR = self._CR.copy()
        mask = np.random.random(size=self._pop_size) < self._t_CR

        random_values = np.random.random(size=np.sum(mask))
        mutate_CR[mask] = random_values
        return mutate_CR

    def _get_new_population(self: jDE) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
        )

        mutate_F = self._get_mutate_F()
        mutate_CR = self._get_mutate_CR()

        mutant_cr_b_g = np.array(
            [
                get_new_individ_g(individ_g=self._population_g_i[i], F=mutate_F[i], CR=mutate_CR[i])
                for i in range(self._pop_size)
            ],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]
        self._F[mask] = mutate_F[mask]
        self._CR[mask] = mutate_CR[mask]

    def _update_data(self: jDE) -> None:
        super()._update_data()
        self._update_stats(F=self._F, CR=self._CR)



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base._ea import EvolutionaryAlgorithm
from ..tools import donothing
from ..tools.operators import best_1
from ..tools.operators import best_2
from ..tools.operators import binomial
from ..tools.operators import current_to_best_1
from ..tools.operators import rand_1
from ..tools.operators import rand_2
from ..tools.operators import rand_to_best1
from ..tools.random import float_population
from ..tools.transformations import bounds_control


class DifferentialEvolution(EvolutionaryAlgorithm):
    """Storn, Rainer & Price, Kenneth. (1995). Differential Evolution: A Simple and Efficient
    Adaptive Scheme for Global Optimization Over Continuous Spaces.
    Journal of Global Optimization. 23"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        left: NDArray[np.float64],
        right: NDArray[np.float64],
        mutation: str = "rand_1",
        F: float = 0.5,
        CR: float = 0.5,
        elitism: bool = True,
        init_population: Optional[NDArray[np.float64]] = None,
        genotype_to_phenotype: Callable[[NDArray[np.float64]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        self._left: NDArray[np.float64] = left
        self._right: NDArray[np.float64] = right
        self._specified_mutation: str = mutation
        self._F: Union[float, NDArray[np.float64]] = F
        self._CR: Union[float, NDArray[np.float64]] = CR

        self._mutation_pool: Dict[str, Callable] = {
            "best_1": best_1,
            "rand_1": rand_1,
            "current_to_best_1": current_to_best_1,
            "rand_to_best1": rand_to_best1,
            "best_2": best_2,
            "rand_2": rand_2,
        }

    def _get_init_population(self: DifferentialEvolution) -> None:
        if self._init_population is None:
            self._population_g_i = float_population(
                pop_size=self._pop_size, left=self._left, right=self._right
            )
        else:
            self._population_g_i = self._init_population.copy()

        self._population_ph_i = self._get_phenotype(self._population_g_i)
        self._fitness_i = self._get_fitness(self._population_ph_i)

    def _get_new_individ_g(
        self: DifferentialEvolution,
        individ_g: NDArray[np.float64],
        F: float,
        CR: float,
    ) -> NDArray[np.float64]:
        mutation_func = self._mutation_pool[self._specified_mutation]

        mutant_g = mutation_func(
            individ_g, self._thefittest._genotype, self._population_g_i, np.float64(F)
        )

        mutant_cr_g = binomial(individ_g, mutant_g, np.float64(CR))
        return bounds_control(mutant_cr_g, self._left, self._right)

    def _get_new_population(self: DifferentialEvolution) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
            F=self._F,
            CR=self._CR,
        )

        mutant_cr_b_g = np.array(
            [get_new_individ_g(individ_g=self._population_g_i[i]) for i in range(self._pop_size)],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]

    def _from_population_g_to_fitness(self: EvolutionaryAlgorithm) -> None:
        self._update_data()

        if self._elitism:
            (
                self._population_g_i[-1],
                self._population_ph_i[-1],
                self._fitness_i[-1],
            ) = self._thefittest.get().values()

        self._adapt()



from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from numba import float64
from numba import njit

import numpy as np
from numpy.typing import NDArray

from ..tools import find_end_subtree_from_i
from ..tools import find_first_difference_between_two


def common_region(trees: Union[List, NDArray]) -> Tuple:
    terminate = False
    indexes = []
    common_indexes: List[List[int]] = []
    border_indexes: List[List[int]] = []
    for tree in trees:
        indexes.append(list(range(len(tree._nodes))))
        common_indexes.append([])
        border_indexes.append([])

    while not terminate:
        inner_break = False
        iters = min(list(map(len, indexes)))

        for i in range(iters):
            first_n_args = trees[0]._nodes[indexes[0][i]]._n_args
            common_indexes[0].append(indexes[0][i])
            for j in range(1, len(indexes)):
                common_indexes[j].append(indexes[j][i])
                if first_n_args != trees[j]._nodes[indexes[j][i]]._n_args:
                    inner_break = True

            if inner_break:
                for j in range(0, len(indexes)):
                    border_indexes[j].append(indexes[j][i])
                break

        for j in range(len(indexes)):
            _, right = trees[j].subtree_id(common_indexes[j][-1])
            delete_to = indexes[j].index(right - 1) + 1
            indexes[j] = indexes[j][delete_to:]

            if len(indexes[j]) < 1:
                terminate = True
                break

    return common_indexes, border_indexes


def common_region_two_trees(
    n_args_array_1: NDArray[np.int64], n_args_array_2: NDArray[np.int64]
) -> Tuple:
    index_list_1 = range(len(n_args_array_1))
    index_list_2 = range(len(n_args_array_2))
    index_1 = np.int64(0)
    index_2 = np.int64(0)

    common_1: List[List[int]] = []
    common_2: List[List[int]] = []
    border_1 = []
    border_2 = []

    while True:
        if index_1 < len(n_args_array_1) and index_2 < len(n_args_array_2):
            id_1 = index_1
            id_2 = index_2
            end = find_first_difference_between_two(n_args_array_1[id_1:], n_args_array_2[id_2:])
            index_1 = index_1 + end
            index_2 = index_2 + end
            common_1.extend(index_list_1[id_1 : index_1 + 1])
            common_2.extend(index_list_2[id_2 : index_2 + 1])

        if len(n_args_array_1) - 1 > index_1 or len(n_args_array_2) - 1 > index_2:
            border_1.append(index_1)
            border_2.append(index_2)
            index_1 = find_end_subtree_from_i(index_1, n_args_array_1)
            index_2 = find_end_subtree_from_i(index_2, n_args_array_2)
        else:
            break

    return [common_1, common_2], [border_1, border_2]


def common_region_(trees: Union[List, NDArray]) -> Tuple:
    if len(trees) == 2:
        to_return = common_region_two_trees(trees[0]._n_args, trees[1]._n_args)
    else:
        to_return = common_region_(trees)

    return to_return


def numpy_group_by(group: NDArray, by: NDArray) -> Tuple:
    argsort = np.argsort(by)
    group = group[argsort]
    by = by[argsort]

    keys, cut_index = np.unique(by, return_index=True)
    groups = np.split(group, cut_index)[1:]
    return keys, groups


def lehmer_mean(
    x: NDArray[np.float64], power: int = 2, weight: Optional[NDArray[np.float64]] = None
) -> float:
    weight_arg: Union[NDArray[np.float64], int]
    if weight is None:
        weight_arg = 1
    else:
        weight_arg = weight

    x_up = weight_arg * np.power(x, power)
    x_down = weight_arg * np.power(x, power - 1)
    return np.sum(x_up) / np.sum(x_down)


def rank_data(arr: NDArray[Union[np.int64, np.float64]]) -> NDArray[np.float64]:
    arange = np.arange(len(arr), dtype=np.int64)

    argsort = np.argsort(arr)
    arr_sorted = arr.copy()[argsort]

    cond = np.r_[True, arr_sorted[1:] != arr_sorted[:-1]]
    raw_ranks = np.r_[arange[cond == True], len(arange)]
    ranks = (raw_ranks[1:] + raw_ranks[:-1] + 1) / 2
    count = raw_ranks[1:] - raw_ranks[:-1]

    results = np.empty_like(arr, dtype=np.float64)
    results[argsort] = ranks.repeat(count)
    return results


@njit(float64[:](float64[:]))
def protect_norm(x: NDArray[np.float64]) -> NDArray[np.float64]:
    result = np.empty(len(x), dtype=np.float64)
    sum_ = np.sum(x, dtype=np.float64)
    if sum_ > 0:
        for i in range(result.size):
            result[i] = x[i] / sum_
    else:
        value = 1 / len(x)
        for i in range(result.size):
            result[i] = value
    return result


def scale_data(arr: NDArray[Union[np.int64, np.float64]]) -> NDArray[np.float64]:
    arr = arr.copy()
    max_ = arr.max()
    min_ = arr.min()
    if max_ == min_:
        to_return = np.ones_like(arr, dtype=np.float64)
    else:
        to_return = ((arr - min_) / (max_ - min_)).astype(np.float64)
    return to_return


def numpy_bit_to_int(
    bit_array: NDArray[np.int64], powers: Optional[NDArray[np.int64]] = None
) -> NDArray[np.int64]:
    if powers is None:
        powers = 2 ** np.arange(bit_array.shape[1], dtype=np.byte)
    arange_ = powers[: bit_array.shape[1]][::-1]
    int_array = np.sum(bit_array * arange_, axis=1)
    return int_array


def numpy_int_to_bit(int_array: NDArray[np.int64]) -> NDArray[np.byte]:
    result = []
    bit = int_array % 2
    remains = int_array // 2
    result.append(bit)
    while np.sum(remains) > 0:
        bit = remains % 2
        remains = remains // 2
        result.append(bit)
    bit_array = np.array(result, dtype=np.int8)[::-1].T
    return bit_array


def numpy_gray_to_bit(gray_array: NDArray[np.byte]) -> NDArray[np.byte]:
    bit_array = np.logical_xor.accumulate(gray_array, axis=-1).astype(np.byte)
    return bit_array


def numpy_bit_to_gray(bit_array: NDArray[np.byte]) -> NDArray[np.byte]:
    cut_gray = np.logical_xor(bit_array[:, :-1], bit_array[:, 1:])
    gray_array = np.hstack([bit_array[:, 0].reshape(-1, 1), cut_gray])
    return gray_array


class SamplingGrid:
    def __init__(self, fit_by: str = "h") -> None:
        self._fit_by = fit_by
        self.left: NDArray[np.float64]
        self.right: NDArray[np.float64]
        self.parts: NDArray[np.int64]
        self.h: NDArray[np.float64]
        self._power_arange: NDArray[np.int64]

    def _culc_h_from_parts(
        self, left: NDArray[np.float64], right: NDArray[np.float64], parts: NDArray[np.int64]
    ) -> NDArray[np.float64]:
        h = (right - left) / (2.0**parts - 1)
        return h

    def _culc_parts_from_h(
        self, left: NDArray[np.float64], right: NDArray[np.float64], h: NDArray[np.float64]
    ) -> NDArray[np.int64]:
        parts = np.ceil(np.log2((right - left) / h + 1)).astype(int)
        return parts

    def _decode(self, bit_array_i: NDArray[np.byte]) -> NDArray[np.int64]:
        int_convert = numpy_bit_to_int(bit_array_i, self._power_arange)
        return int_convert

    def fit(
        self,
        left: NDArray[np.float64],
        right: NDArray[np.float64],
        arg: Union[NDArray[np.float64], NDArray[np.int64]],
    ):
        self.left = left
        self.right = right

        assert self._fit_by in ["h", "parts"], f"incorrect option {self._fit_by} for fit_by."
        "The available ones are 'h' and 'parts'"
        if self._fit_by == "h":
            min_h = arg
            self.parts = self._culc_parts_from_h(left, right, min_h)
            self.h = self._culc_h_from_parts(left, right, self.parts)
        else:
            self.parts = arg
            self.h = self._culc_h_from_parts(left, right, self.parts)

        self._power_arange = 2 ** np.arange(self.parts.max(), dtype=np.int64)
        return self

    def transform(self, population: np.ndarray) -> np.ndarray:
        splits = np.add.accumulate(self.parts)
        p_parts = np.split(population, splits[:-1], axis=1)

        int_array = np.array(list(map(self._decode, p_parts))).T
        float_array = self.left[np.newaxis, :] + self.h[np.newaxis, :] * int_array
        return float_array

    def _float_to_bit(self, float_array: np.ndarray, left: np.ndarray, h: np.ndarray) -> np.ndarray:
        grid_number = (float_array - left) / h
        int_array = np.rint(grid_number)
        bit_array = numpy_int_to_bit(int_array)
        return bit_array

    def inverse_transform(self, population: np.ndarray) -> np.ndarray:
        map_ = map(self._float_to_bit, population.T, self.left, self.h)
        bit_array = np.hstack(list(map_))
        return bit_array


class GrayCode(SamplingGrid):
    def __init__(self, fit_by: str = "h") -> None:
        SamplingGrid.__init__(self, fit_by)

    def _decode(self, gray_array_i: np.ndarray) -> np.ndarray:
        bit_array_i = numpy_gray_to_bit(gray_array_i)
        int_convert = numpy_bit_to_int(bit_array_i, self._power_arange)
        return int_convert

    def _float_to_bit(self, float_array: np.ndarray, left: np.ndarray, h: np.ndarray) -> np.ndarray:
        grid_number = (float_array - left) / h
        int_array = np.rint(grid_number)
        bit_array = numpy_int_to_bit(int_array)
        gray_array = numpy_bit_to_gray(bit_array)
        return gray_array


@njit(float64[:](float64[:], float64[:], float64[:]))
def bounds_control(
    array: NDArray[np.float64], left: NDArray[np.float64], right: NDArray[np.float64]
) -> NDArray[np.float64]:
    to_return = array.copy()

    size = len(array)
    for i in range(size):
        if array[i] < left[i]:
            to_return[i] = left[i]
        elif array[i] > right[i]:
            to_return[i] = right[i]
    return to_return


@njit(float64[:](float64[:], float64[:], float64[:]))
def bounds_control_mean(
    array: NDArray[np.float64], left: NDArray[np.float64], right: NDArray[np.float64]
) -> NDArray[np.float64]:
    to_return = array.copy()
    size = len(array)
    for i in range(size):
        if array[i] < left[i]:
            to_return[i] = (left[i] + array[i]) / 2
        elif array[i] > right[i]:
            to_return[i] = (right[i] + array[i]) / 2
    return to_return



from numba import boolean
from numba import float64
from numba import int64
from numba import njit

import numpy as np
from numpy.typing import NDArray


@njit(int64(int64, int64[:]))
def find_end_subtree_from_i(index: np.int64, n_args_array: NDArray[np.int64]) -> np.int64:
    n_index = index + 1
    possible_steps = n_args_array[index]
    while possible_steps:
        possible_steps += n_args_array[n_index] - 1
        n_index += 1
    return n_index


@njit(int64[:](int64, int64[:]))
def find_id_args_from_i(index: np.int64, n_args_array: NDArray[np.int64]) -> NDArray[np.int64]:
    out = np.empty(n_args_array[index], dtype=np.int64)
    out[0] = index + 1
    for i in range(1, len(out)):
        next_stop = find_end_subtree_from_i(out[i - 1], n_args_array)
        out[i] = next_stop
    return out


@njit(int64[:](int64, int64[:]))
def get_levels_tree_from_i(origin: np.int64, n_args_array: NDArray[np.int64]) -> NDArray[np.int64]:
    d_i = -1
    s = [1]
    d = [-1]
    result_list = []
    for n_arg in n_args_array[origin:]:
        s[-1] = s[-1] - 1
        if s[-1] == 0:
            s.pop()
            d_i = d.pop() + 1
        else:
            d_i = d[-1] + 1
        result_list.append(d_i)
        if n_arg > 0:
            s.append(n_arg)
            d.append(d_i)
        if len(s) == 0:
            break
    return np.array(result_list, dtype=np.int64)


@njit(int64(int64[:], int64[:]))
def find_first_difference_between_two(
    array_1: NDArray[np.int64], array_2: NDArray[np.int64]
) -> np.int64:
    for i in np.arange(min(len(array_1), len(array_2)), dtype=np.int64):
        if array_1[i] != array_2[i]:
            break
    return i


@njit(int64(float64, float64[:]))
def binary_search_interval(value: np.float64, intervals: NDArray[np.float64]) -> int:
    if value <= intervals[0]:
        ind = 0
    else:
        left = 0
        right = len(intervals) - 1
        while right - left > 1:
            mid = (left + right) // 2
            if value <= intervals[mid]:
                right = mid
            else:
                left = mid
        ind = right
    return ind


@njit(boolean(int64, int64[:], int64))
def check_for_value(value: np.int64, index_array: NDArray[np.int64], end: np.int64) -> bool:
    found = False
    for i in range(end):
        if value == index_array[i]:
            found = True
            break
    return found


@njit(int64[:](float64[:], int64))
def argsort_k(array: NDArray[np.float64], k: np.int64) -> NDArray[np.int64]:
    size = len(array)
    array_copy = array.copy()
    to_return = np.arange(size, dtype=np.int64)
    for i in range(k):
        max_ = array_copy[i]
        for j in range(i, size):
            if array_copy[j] > max_:
                max_ = array_copy[j]
                max_id = j
        array_copy[i], array_copy[max_id] = array_copy[max_id], array_copy[i]
        to_return[i], to_return[max_id] = to_return[max_id], to_return[i]
    return to_return


@njit(int64[:](float64[:], float64))
def find_pbest_id(array: NDArray[np.float64], p: np.float64) -> NDArray[np.int64]:
    size = len(array)
    count = max(np.int64(1), np.int64(p * size))
    argsort = argsort_k(array, count)
    to_return = argsort[:count]
    return to_return


@njit(float64[:](float64[:, :]))
def max_axis(array: NDArray[np.float64]) -> NDArray[np.float64]:
    res = np.zeros((array.shape[1]), dtype=np.float64)
    for i in range(array.shape[1]):
        res[i] = np.max(array[:, i])
    return res



from typing import Any

from ..base._net import Net
from ..base import Tree


def print_net(net: Net, ax: Any = None) -> None:
    import networkx as nx

    graph = net.get_graph()
    G = nx.Graph()
    G.add_nodes_from(graph["nodes"])
    G.add_edges_from(graph["connects"])

    connects_label = {}
    for i, connects_i in enumerate(graph["connects"]):
        connects_label[tuple(connects_i)] = i

    nx.draw_networkx_nodes(
        G,
        pos=graph["positions"],
        node_color=graph["colors"],
        edgecolors="black",
        linewidths=0.5,
        ax=ax,
    )
    nx.draw_networkx_edges(
        G, pos=graph["positions"], style="-", edge_color=graph["weights_colors"], ax=ax, alpha=0.5
    )

    nx.draw_networkx_labels(G, graph["positions"], graph["labels"], font_size=10, ax=ax)


def print_tree(tree: Tree, ax: Any = None) -> None:
    import networkx as nx

    graph = tree.get_graph(False)

    G = nx.Graph()
    G.add_nodes_from(graph["nodes"])
    G.add_edges_from(graph["edges"])

    nx.draw_networkx_nodes(
        G, graph["pos"], node_color=graph["colors"], edgecolors="black", linewidths=0.5, ax=ax
    )
    nx.draw_networkx_edges(G, graph["pos"], style="-", ax=ax)
    nx.draw_networkx_labels(G, graph["pos"], graph["labels"], font_size=10, ax=ax)



import os
from typing import Callable
from typing import Tuple

import numpy as np
from numpy.typing import NDArray

path = os.path.dirname(__file__) + "/_data/"

fbias_data = np.loadtxt(path + "fbias_data.txt")
# 1
sphere_func_data = np.loadtxt(path + "sphere_func_data.txt")
# 2
schwefel_102_data = np.loadtxt(path + "schwefel_102_data.txt")
# 3
high_cond_elliptic_rot_data = np.loadtxt(path + "high_cond_elliptic_rot_data.txt")[:50]
elliptic_m_d2 = np.loadtxt(path + "elliptic_M_D2.txt")
elliptic_m_d10 = np.loadtxt(path + "elliptic_M_D10.txt")
elliptic_m_d30 = np.loadtxt(path + "elliptic_M_D30.txt")
elliptic_m_d50 = np.loadtxt(path + "elliptic_M_D50.txt")
# 5
schwefel_206_data = np.loadtxt(path + "schwefel_206_data.txt")
o_206 = schwefel_206_data[0]
A_206 = schwefel_206_data[1:]
# 6
rosenbrock_func_data = np.loadtxt(path + "rosenbrock_func_data.txt")
# 7
griewank_func_data = np.loadtxt(path + "griewank_func_data.txt")[:50]
griewank_m_d2 = np.loadtxt(path + "griewank_M_D2.txt")
griewank_m_d10 = np.loadtxt(path + "griewank_M_D10.txt")
griewank_m_d30 = np.loadtxt(path + "griewank_M_D30.txt")
griewank_m_d50 = np.loadtxt(path + "griewank_M_D50.txt")
# 8
ackley_func_data = np.loadtxt(path + "ackley_func_data.txt")[:50]
ackley_m_d2 = np.loadtxt(path + "ackley_M_D2.txt")
ackley_m_d10 = np.loadtxt(path + "ackley_M_D10.txt")
ackley_m_d30 = np.loadtxt(path + "ackley_M_D30.txt")
ackley_m_d50 = np.loadtxt(path + "ackley_M_D50.txt")
# 9-10
rastrigin_func_data = np.loadtxt(path + "rastrigin_func_data.txt")[:50]
rastrigin_m_d2 = np.loadtxt(path + "rastrigin_M_D2.txt")
rastrigin_m_d10 = np.loadtxt(path + "rastrigin_M_D10.txt")
rastrigin_m_d30 = np.loadtxt(path + "rastrigin_M_D30.txt")
rastrigin_m_d50 = np.loadtxt(path + "rastrigin_M_D50.txt")
# 11
weierstrass_m_d2 = np.loadtxt(path + "weierstrass_M_D2.txt")
weierstrass_m_d10 = np.loadtxt(path + "weierstrass_M_D10.txt")
weierstrass_m_d30 = np.loadtxt(path + "weierstrass_M_D30.txt")
weierstrass_m_d50 = np.loadtxt(path + "weierstrass_M_D50.txt")
weierstrass_data = np.loadtxt(path + "weierstrass_data.txt")[:50]
# 12
schwefel_213_data = np.loadtxt(path + "schwefel_213_data.txt")
# 13
EF8F2_func_data = np.loadtxt(path + "EF8F2_func_data.txt")
a_213 = schwefel_213_data[:100]
b_213 = schwefel_213_data[100:200]
alpha_213 = schwefel_213_data[-1]
# 14
E_ScafferF6_func_data = np.loadtxt(path + "E_ScafferF6_func_data.txt")[:50]
E_ScafferF6_m_d2 = np.loadtxt(path + "E_ScafferF6_M_D2.txt")
E_ScafferF6_m_d10 = np.loadtxt(path + "E_ScafferF6_M_D10.txt")
E_ScafferF6_m_d30 = np.loadtxt(path + "E_ScafferF6_M_D30.txt")
E_ScafferF6_m_d50 = np.loadtxt(path + "E_ScafferF6_M_D50.txt")
# 15
hybrid_func1_data = np.loadtxt(path + "hybrid_func1_data.txt")
# 16
hybrid_func1_m_d2 = np.loadtxt(path + "hybrid_func1_M_D2.txt")
hybrid_func1_m_d10 = np.loadtxt(path + "hybrid_func1_M_D10.txt")
hybrid_func1_m_d30 = np.loadtxt(path + "hybrid_func1_M_D30.txt")
hybrid_func1_m_d50 = np.loadtxt(path + "hybrid_func1_M_D50.txt")
# 18
hybrid_func2_data = np.loadtxt(path + "hybrid_func2_data.txt")
hybrid_func2_m_d2 = np.loadtxt(path + "hybrid_func2_M_D2.txt")
hybrid_func2_m_d10 = np.loadtxt(path + "hybrid_func2_M_D10.txt")
hybrid_func2_m_d30 = np.loadtxt(path + "hybrid_func2_M_D30.txt")
hybrid_func2_m_d50 = np.loadtxt(path + "hybrid_func2_M_D50.txt")
# 21
hybrid_func3_data = np.loadtxt(path + "hybrid_func3_data.txt")
hybrid_func3_m_d2 = np.loadtxt(path + "hybrid_func3_M_D2.txt")
hybrid_func3_m_d10 = np.loadtxt(path + "hybrid_func3_M_D10.txt")
hybrid_func3_m_d30 = np.loadtxt(path + "hybrid_func3_M_D30.txt")
hybrid_func3_m_d50 = np.loadtxt(path + "hybrid_func3_M_D50.txt")
# 22
hybrid_func3_hm_d2 = np.loadtxt(path + "hybrid_func3_HM_D2.txt")
hybrid_func3_hm_d10 = np.loadtxt(path + "hybrid_func3_HM_D10.txt")
hybrid_func3_hm_d30 = np.loadtxt(path + "hybrid_func3_HM_D30.txt")
hybrid_func3_hm_d50 = np.loadtxt(path + "hybrid_func3_HM_D50.txt")
# 24
hybrid_func4_data = np.loadtxt(path + "hybrid_func4_data.txt")
hybrid_func4_m_d2 = np.loadtxt(path + "hybrid_func4_M_D2.txt")
hybrid_func4_m_d10 = np.loadtxt(path + "hybrid_func4_M_D10.txt")
hybrid_func4_m_d30 = np.loadtxt(path + "hybrid_func4_M_D30.txt")
hybrid_func4_m_d50 = np.loadtxt(path + "hybrid_func4_M_D50.txt")


class TestFunction:
    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.f(x)

    def build_grid(self, x: NDArray[np.float64], y: NDArray[np.float64]) -> NDArray[np.float64]:
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate([x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class TestShiftedFunction:
    def __init__(self, fbias: NDArray[np.float64], x_shift: NDArray[np.float64]) -> None:
        self.fbias = fbias
        self.x_shift = x_shift

    def shift(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = x.shape
        axis = [1] * (len(shape) - 1) + [-1]
        return x - self.x_shift[: shape[-1]].reshape(axis)

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        z = self.shift(x)
        return self.f(z) + self.fbias

    def build_grid(self, x: NDArray[np.float64], y: NDArray[np.float64]) -> NDArray[np.float64]:
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate([x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class TestShiftedRotatedFunction:
    def __init__(
        self,
        fbias: NDArray[np.float64],
        x_shift: NDArray[np.float64],
        rotate_M_D2: NDArray[np.float64],
        rotate_M_D10: NDArray[np.float64],
        rotate_M_D30: NDArray[np.float64],
        rotate_M_D50: NDArray[np.float64],
    ) -> None:
        self.fbias = fbias
        self.x_shift = x_shift
        self.rotate_M_D2 = rotate_M_D2
        self.rotate_M_D10 = rotate_M_D10
        self.rotate_M_D30 = rotate_M_D30
        self.rotate_M_D50 = rotate_M_D50

    def shift(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = x.shape
        axis = [1] * (len(shape) - 1) + [-1]
        return x - self.x_shift[: shape[-1]].reshape(axis)

    def rotate(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.shape[1] == 2:
            z = x @ self.rotate_M_D2
        elif x.shape[1] == 10:
            z = x @ self.rotate_M_D10
        elif x.shape[1] == 30:
            z = x @ self.rotate_M_D30
        elif x.shape[1] == 50:
            z = x @ self.rotate_M_D50
        return z

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        z = self.shift(x)
        z_rotated = self.rotate(z)
        return self.f(z_rotated) + self.fbias

    def build_grid(self, x: NDArray[np.float64], y: NDArray[np.float64]) -> NDArray[np.float64]:
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate([x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class SampleHybridCompositionFunction:
    def __init__(
        self,
        basic_functions: Tuple,
        sigmas: NDArray[np.float64],
        lambdas: NDArray[np.float64],
        fbias: NDArray[np.float64],
        x_shift: NDArray[np.float64],
        M_D2: NDArray[np.float64],
        M_D10: NDArray[np.float64],
        M_D30: NDArray[np.float64],
        M_D50: NDArray[np.float64],
    ) -> None:
        self.basic_functions = basic_functions
        self.sigmas = sigmas
        self.lambdas = lambdas
        self.M_D2 = np.split(M_D2, range(M_D2.shape[1], M_D2.shape[0], M_D2.shape[1]))
        self.M_D10 = np.split(M_D10, range(M_D10.shape[1], M_D10.shape[0], M_D10.shape[1]))
        self.M_D30 = np.split(M_D30, range(M_D30.shape[1], M_D30.shape[0], M_D30.shape[1]))
        self.M_D50 = np.split(M_D50, range(M_D50.shape[1], M_D50.shape[0], M_D50.shape[1]))
        self.fbias = fbias
        self.x_shift = x_shift

        self.bias = np.array([0, 100, 200, 300, 400, 500, 600, 700, 800, 900], dtype=np.float64)
        self.C = 2000

    def shift(self, x: NDArray[np.float64], i: np.int64) -> NDArray[np.float64]:
        shape = x.shape
        axis = [1] * (len(shape) - 1) + [-1]
        return x - self.x_shift[i][: shape[-1]].reshape(axis)

    def rotate(
        self,
        x: NDArray[np.float64],
        M_D2: NDArray[np.float64],
        M_D10: NDArray[np.float64],
        M_D30: NDArray[np.float64],
        M_D50: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if x.shape[1] == 2:
            z = x @ M_D2
        elif x.shape[1] == 10:
            z = x @ M_D10
        elif x.shape[1] == 30:
            z = x @ M_D30
        elif x.shape[1] == 50:
            z = x @ M_D50
        return z

    def calc_w(
        self,
        x: NDArray[np.float64],
        shift_i: np.float64,
        sigma_i: np.float64,
        i: np.int64,
    ) -> Tuple:
        D = x.shape[1]
        shift_x = self.shift(x, i)

        up = np.sum(shift_x**2, axis=-1)
        down = 2 * D * (sigma_i * sigma_i)

        w_i = np.exp(-(up / down))

        return w_i, shift_x

    def procces_i_function(
        self,
        x: NDArray[np.float64],
        func_i: Callable,
        sigma_i: np.float64,
        lambda_i: np.float64,
        shift_i: np.float64,
        M_D2i: NDArray[np.float64],
        M_D10i: NDArray[np.float64],
        M_D30i: NDArray[np.float64],
        M_D50i: NDArray[np.float64],
        i: np.int64,
    ) -> Tuple:
        y = np.full(shape=x.shape, fill_value=5)
        w_i, shift_x = self.calc_w(x, shift_i, sigma_i, i)

        arg1 = self.rotate(shift_x / lambda_i, M_D2i, M_D10i, M_D30i, M_D50i)
        fit_i = func_i()(arg1)

        arg2 = self.rotate(y / lambda_i, M_D2i, M_D10i, M_D30i, M_D50i)
        f_max_i = func_i()(arg2)

        fit_i = self.C * fit_i / f_max_i

        return w_i, fit_i

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        result = tuple(
            self.procces_i_function(
                x, func_i, sigma_i, lambda_i, shift_i, M_D2i, M_D10i, M_D30i, M_D50i, i
            )
            for func_i, sigma_i, lambda_i, shift_i, M_D2i, M_D10i, M_D30i, M_D50i, i in zip(
                self.basic_functions,
                self.sigmas,
                self.lambdas,
                self.x_shift,
                self.M_D2,
                self.M_D10,
                self.M_D30,
                self.M_D50,
                range(len(self.M_D50)),
            )
        )
        w_i, fit_i = list(zip(*result))
        w_i = np.array(w_i, np.float64).T
        fit_i = np.array(fit_i, np.float64).T
        max_w = np.max(w_i, axis=1)

        for i, w_i_j in enumerate(w_i.T):
            cond = w_i_j != max_w
            w_i[:, i][cond] = w_i[:, i][cond] * (1.0 - max_w[cond] ** 10)

        w_i = w_i / np.sum(w_i, axis=-1)[:, np.newaxis]

        return np.sum(w_i * (fit_i + self.bias), axis=-1) + self.fbias

    def build_grid(self, x: NDArray[np.float64], y: NDArray[np.float64]) -> NDArray[np.float64]:
        x1, y1 = np.meshgrid(x, y)
        xy = np.concatenate([x1[:, :, np.newaxis], y1[:, :, np.newaxis]], axis=2)
        z = np.zeros(shape=xy.shape[:-1])
        for i, x_i in enumerate(xy):
            z[i] = self(x_i)
        return z


class OneMax(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.sum(x, axis=1, dtype=np.float64)


class Sphere(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.sum(x**2, axis=-1)


class Schwefe1_2(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.sum(np.add.accumulate(x, axis=-1) ** 2, axis=-1)


class HighConditionedElliptic(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        i = np.arange(1, x.shape[1] + 1)
        demension = x.shape[1]
        return np.sum((1e6 ** ((i - 1) / (demension - 1))) * x**2, axis=-1)


class Rosenbrock(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        value = 100 * ((x.T[:-1] ** 2 - x.T[1:]) ** 2) + (x.T[:-1] - 1) ** 2
        return np.sum(value.T, axis=-1)


class Rastrigin(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.sum(x**2 - 10 * np.cos(2 * np.pi * x) + 10, axis=-1)


class Griewank(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        sqrt_i = np.sqrt(np.arange(1, x.shape[1] + 1))
        sum_ = np.sum((x**2) / 4000, axis=-1)
        prod_ = np.prod(np.cos(x / sqrt_i), axis=-1)
        return sum_ - prod_ + 1


class Ackley(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        a = 20
        c = 2 * np.pi
        b = 0.2
        D = x.shape[1]
        left = -a * np.exp(-b * np.sqrt(np.sum(x**2, axis=1) / D))
        right = -np.exp((1 / D) * np.sum(np.cos(c * x), axis=1))
        return left + right + a + np.exp(1)


class Weierstrass(TestFunction):
    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        a = 0.5
        b = 3
        k_max = np.arange(20, dtype=np.int64)
        D = x.shape[1]

        two_pi_power_b = 2 * np.pi * (b**k_max)
        a_power_k = a**k_max
        arg1 = two_pi_power_b * (x[:, :, np.newaxis] + 0.5)
        arg2 = two_pi_power_b * 0.5
        left = a_power_k * np.cos(arg1)
        right = a_power_k * np.cos(arg2)

        left_sum = np.sum(np.sum(left, axis=-1), axis=-1)
        right_sum = np.sum(right, axis=-1)
        return left_sum - D * right_sum


class F8F2(TestFunction):
    def __init__(self) -> None:
        self.rosenbrock_f = Rosenbrock()
        self.griewank_f = Griewank()

    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.shape[1] == 2:
            indexes = np.array([0, 1, 1, 0], dtype=np.int64)
        else:
            indexes = np.kron(np.arange(1, x.shape[1] - 1, dtype=np.int64), np.array([1, 1]))

            indexes = np.insert(indexes, [0], [0])

            indexes = np.append(indexes, np.array([x.shape[1] - 1, x.shape[1] - 1, 0]))

        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F2 = self.rosenbrock_f(vertical_X)
        F8 = self.griewank_f(F2.reshape(-1, 1))
        horizontal_X = F8.reshape(x.shape[0], -1)
        return np.sum(horizontal_X, axis=-1)


class ExpandedScaffers_F6(TestFunction):
    def Scaffes_F6(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        sum_of_power = x[:, 0] ** 2 + x[:, 1] ** 2
        up = np.sin(np.sqrt(sum_of_power)) ** 2 - 0.5
        down = (1 + 0.001 * (sum_of_power)) ** 2
        return 0.5 + up / down

    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.shape[1] == 2:
            indexes = np.array([0, 1, 1, 0], dtype=np.int64)
        else:
            indexes = np.kron(np.arange(1, x.shape[1] - 1, dtype=np.int64), np.array([1, 1]))

            indexes = np.insert(indexes, [0], [0])

            indexes = np.append(indexes, np.array([x.shape[1] - 1, x.shape[1] - 1, 0]))
        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F = self.Scaffes_F6(vertical_X)
        horizontal_X = F.reshape(x.shape[0], -1)

        return np.sum(horizontal_X, axis=-1)


class NonContinuosRastrigin(Rastrigin):
    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        cond = np.abs(x) >= 0.5
        x[cond] = (np.round(2 * x) / 2)[cond]
        return super().__call__(x)


class NonContinuosExpandedScaffers_F6(ExpandedScaffers_F6):
    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        cond = np.abs(x) >= 0.5
        x[cond] = (np.round(2 * x) / 2)[cond]
        return super().__call__(x)


class SphereWithNoise(Sphere):
    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        y = super().__call__(x)
        noise = 1 + 0.1 * np.abs(np.random.normal(size=y.shape))
        return y * noise


# CEC05 #1
class ShiftedSphere(TestShiftedFunction, Sphere):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[0], x_shift=sphere_func_data)


# CEC05 #2
class ShiftedSchwefe1_2(TestShiftedFunction, Schwefe1_2):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[1], x_shift=schwefel_102_data)


# CEC05 #3
class ShiftedRotatedHighConditionedElliptic(TestShiftedRotatedFunction, HighConditionedElliptic):
    def __init__(self) -> None:
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[2],
            x_shift=high_cond_elliptic_rot_data,
            rotate_M_D2=elliptic_m_d2,
            rotate_M_D10=elliptic_m_d10,
            rotate_M_D30=elliptic_m_d30,
            rotate_M_D50=elliptic_m_d50,
        )


# CEC05 #4
class ShiftedSchwefe1_2WithNoise(TestShiftedFunction, Schwefe1_2):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[3], x_shift=schwefel_102_data)

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        z = self.shift(x)
        y = self.f(z)
        noise = 1 + 0.4 * np.abs(np.random.normal(size=y.shape))
        return y * noise + self.fbias


# CEC05 #5
class Schwefel2_6(TestShiftedFunction):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[4], x_shift=o_206)

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        D = x.shape[1]
        self.x_shift[np.floor(3 * D / 4).astype(np.int64) - 1 : D] = 100
        self.x_shift[: np.ceil(D / 4).astype(np.int64)] = -100
        return self.f(x) + self.fbias

    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        o = self.x_shift[: x.shape[1]]
        A = A_206[: x.shape[1], : x.shape[1]]
        Ax = A @ x.T
        B = A @ o
        fx = np.abs(Ax - B[:, np.newaxis])
        return np.max(fx, axis=0)


# CEC05 #6
class ShiftedRosenbrock(TestShiftedFunction, Rosenbrock):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[5], x_shift=rosenbrock_func_data)

    def shift(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = x.shape
        axis = [1] * (len(shape) - 1) + [-1]
        return x - self.x_shift[: shape[-1]].reshape(axis) + 1


# CEC05 #7
class ShiftedRotatedGriewank(TestShiftedRotatedFunction, Griewank):
    def __init__(self) -> None:
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[6],
            x_shift=griewank_func_data,
            rotate_M_D2=griewank_m_d2,
            rotate_M_D10=griewank_m_d10,
            rotate_M_D30=griewank_m_d30,
            rotate_M_D50=griewank_m_d50,
        )


# CEC05 #8
class ShiftedRotatedAckley(TestShiftedRotatedFunction, Ackley):
    def __init__(self) -> None:
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[7],
            x_shift=ackley_func_data,
            rotate_M_D2=ackley_m_d2,
            rotate_M_D10=ackley_m_d10,
            rotate_M_D30=ackley_m_d30,
            rotate_M_D50=ackley_m_d50,
        )

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        D = x.shape[1]
        self.x_shift[:D:2] = -32.0
        return super().__call__(x)


# CEC05 #9
class ShiftedRastrigin(TestShiftedFunction, Rastrigin):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[8], x_shift=rastrigin_func_data)


# CEC05 #10
class ShiftedRotatedRastrigin(TestShiftedRotatedFunction, Rastrigin):
    def __init__(self) -> None:
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[9],
            x_shift=rastrigin_func_data,
            rotate_M_D2=rastrigin_m_d2,
            rotate_M_D10=rastrigin_m_d10,
            rotate_M_D30=rastrigin_m_d30,
            rotate_M_D50=rastrigin_m_d50,
        )


# CEC05 #11
class ShiftedRotatedWeierstrass(TestShiftedRotatedFunction, Weierstrass):
    def __init__(self) -> None:
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[10],
            x_shift=weierstrass_data,
            rotate_M_D2=weierstrass_m_d2,
            rotate_M_D10=weierstrass_m_d10,
            rotate_M_D30=weierstrass_m_d30,
            rotate_M_D50=weierstrass_m_d50,
        )


# CEC05 #12
class Schwefel2_13(TestShiftedFunction):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[11], x_shift=alpha_213)

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.f(x) + self.fbias

    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        a = a_213[: x.shape[1], : x.shape[1]]
        b = b_213[: x.shape[1], : x.shape[1]]
        alpha = self.x_shift[: x.shape[1]]

        A = np.sum(a * np.sin(alpha)[np.newaxis :,] + b * np.cos(alpha)[np.newaxis :,], axis=1)

        B1 = a[np.newaxis] * np.sin(x[:, np.newaxis])
        B2 = b[np.newaxis] * np.cos(x[:, np.newaxis])
        B = np.sum(B1 + B2, axis=-1)

        return np.sum((A[np.newaxis :,] - B) ** 2, axis=-1)


# CEC05 #13
class ShiftedExpandedGriewankRosenbrock(TestShiftedFunction):
    def __init__(self) -> None:
        TestShiftedFunction.__init__(self, fbias=fbias_data[12], x_shift=EF8F2_func_data)
        self.rosenbrock_f = Rosenbrock()
        self.griewank_f = Griewank()

    def shift(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = x.shape
        axis = [1] * (len(shape) - 1) + [-1]
        return x - self.x_shift[: shape[-1]].reshape(axis) + 1

    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.shape[1] == 2:
            indexes = np.array([0, 1, 1, 0], dtype=np.int64)
        else:
            indexes = np.kron(np.arange(1, x.shape[1] - 1, dtype=np.int64), np.array([1, 1]))

            indexes = np.insert(indexes, [0], [0])

            indexes = np.append(indexes, np.array([x.shape[1] - 1, x.shape[1] - 1, 0]))

        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F2 = self.rosenbrock_f(vertical_X)
        F8 = self.griewank_f(F2.reshape(-1, 1))
        horizontal_X = F8.reshape(x.shape[0], -1)
        return np.sum(horizontal_X, axis=-1)


# CEC05 #14
class ShiftedRotatedExpandedScaffes_F6(TestShiftedRotatedFunction):
    def __init__(self) -> None:
        TestShiftedRotatedFunction.__init__(
            self,
            fbias=fbias_data[13],
            x_shift=E_ScafferF6_func_data,
            rotate_M_D2=E_ScafferF6_m_d2,
            rotate_M_D10=E_ScafferF6_m_d10,
            rotate_M_D30=E_ScafferF6_m_d30,
            rotate_M_D50=E_ScafferF6_m_d50,
        )

    def Scaffes_F6(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        sum_of_power = x[:, 0] ** 2 + x[:, 1] ** 2
        up = np.sin(np.sqrt(sum_of_power)) ** 2 - 0.5
        down = (1 + 0.001 * (sum_of_power)) ** 2
        return 0.5 + up / down

    def f(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        if x.shape[1] == 2:
            indexes = np.array([0, 1, 1, 0], dtype=np.int64)
        else:
            indexes = np.kron(np.arange(1, x.shape[1] - 1, dtype=np.int64), np.array([1, 1]))

            indexes = np.insert(indexes, [0], [0])

            indexes = np.append(indexes, np.array([x.shape[1] - 1, x.shape[1] - 1, 0]))

        x_indexes = x[:, indexes]

        vertical_X = x_indexes.reshape(-1, 2)
        F = self.Scaffes_F6(vertical_X)
        horizontal_X = F.reshape(x.shape[0], -1)

        return np.sum(horizontal_X, axis=-1)


# CEC05 #15
class HybridCompositionFunction1(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Rastrigin,
                Rastrigin,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
                Ackley,
                Ackley,
                Sphere,
                Sphere,
            ),
            sigmas=np.ones(10),
            lambdas=np.array(
                [1, 1, 10, 10, 5 / 60, 5 / 60, 5 / 32, 5 / 32, 5 / 100, 5 / 100], dtype=np.float64
            ),
            fbias=fbias_data[14],
            x_shift=hybrid_func1_data,
            M_D2=np.vstack([np.eye(2) for _ in range(10)]),
            M_D10=np.vstack([np.eye(10) for _ in range(10)]),
            M_D30=np.vstack([np.eye(30) for _ in range(10)]),
            M_D50=np.vstack([np.eye(50) for _ in range(10)]),
        )


# CEC05 #16
class RotatedVersionHybridCompositionFunction1(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Rastrigin,
                Rastrigin,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
                Ackley,
                Ackley,
                Sphere,
                Sphere,
            ),
            sigmas=np.ones(10),
            lambdas=np.array(
                [1, 1, 10, 10, 5 / 60, 5 / 60, 5 / 32, 5 / 32, 5 / 100, 5 / 100], dtype=np.float64
            ),
            fbias=fbias_data[15],
            x_shift=hybrid_func1_data,
            M_D2=hybrid_func1_m_d2,
            M_D10=hybrid_func1_m_d10,
            M_D30=hybrid_func1_m_d30,
            M_D50=hybrid_func1_m_d50,
        )


# CEC05 #17
class RotatedVersionHybridCompositionFunction1Noise(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Rastrigin,
                Rastrigin,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
                Ackley,
                Ackley,
                Sphere,
                Sphere,
            ),
            sigmas=np.ones(10),
            lambdas=np.array(
                [1, 1, 10, 10, 5 / 60, 5 / 60, 5 / 32, 5 / 32, 5 / 100, 5 / 100], dtype=np.float64
            ),
            fbias=fbias_data[16],
            x_shift=hybrid_func1_data,
            M_D2=hybrid_func1_m_d2,
            M_D10=hybrid_func1_m_d10,
            M_D30=hybrid_func1_m_d30,
            M_D50=hybrid_func1_m_d50,
        )

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        G_x = RotatedVersionHybridCompositionFunction1()(x) - fbias_data[15]
        noise = np.abs(np.random.normal(size=G_x.shape))
        return G_x * (1 + 0.2 * noise) + fbias_data[16]


# CEC05 #18
class RotatedHybridCompositionFunction(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Ackley,
                Ackley,
                Rastrigin,
                Rastrigin,
                Sphere,
                Sphere,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
            ),
            sigmas=np.array([1, 2, 1.5, 1.5, 1, 1, 1.5, 1.5, 2, 2], dtype=np.float64),
            lambdas=np.array(
                [
                    2 * (5 / 32),
                    5 / 32,
                    2 * 1,
                    1,
                    2 * (5 / 100),
                    5 / 100,
                    2 * 10,
                    10,
                    2 * (5 / 60),
                    5 / 60,
                ],
                dtype=np.float64,
            ),
            fbias=fbias_data[17],
            x_shift=hybrid_func2_data,
            M_D2=hybrid_func2_m_d2,
            M_D10=hybrid_func2_m_d10,
            M_D30=hybrid_func2_m_d30,
            M_D50=hybrid_func2_m_d50,
        )
        self.x_shift[9] = 0


# CEC05 #19
class RotatedHybridCompositionFunctionNarrowBasin(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Ackley,
                Ackley,
                Rastrigin,
                Rastrigin,
                Sphere,
                Sphere,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
            ),
            sigmas=np.array([0.1, 2, 1.5, 1.5, 1, 1, 1.5, 1.5, 2, 2], dtype=np.float64),
            lambdas=np.array(
                [
                    0.1 * (5 / 32),
                    5 / 32,
                    2 * 1,
                    1,
                    2 * (5 / 100),
                    5 / 100,
                    2 * 10,
                    10,
                    2 * (5 / 60),
                    5 / 60,
                ],
                dtype=np.float64,
            ),
            fbias=fbias_data[17],
            x_shift=hybrid_func2_data,
            M_D2=hybrid_func2_m_d2,
            M_D10=hybrid_func2_m_d10,
            M_D30=hybrid_func2_m_d30,
            M_D50=hybrid_func2_m_d50,
        )
        self.x_shift[9] = 0


# CEC05 #20
class RotatedHybridCompositionFunctionOptimalBounds(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Ackley,
                Ackley,
                Rastrigin,
                Rastrigin,
                Sphere,
                Sphere,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
            ),
            sigmas=np.array([1, 2, 1.5, 1.5, 1, 1, 1.5, 1.5, 2, 2], dtype=np.float64),
            lambdas=np.array(
                [
                    2 * (5 / 32),
                    5 / 32,
                    2 * 1,
                    1,
                    2 * (5 / 100),
                    5 / 100,
                    2 * 10,
                    10,
                    2 * (5 / 60),
                    5 / 60,
                ],
                dtype=np.float64,
            ),
            fbias=fbias_data[19],
            x_shift=hybrid_func2_data,
            M_D2=hybrid_func2_m_d2,
            M_D10=hybrid_func2_m_d10,
            M_D30=hybrid_func2_m_d30,
            M_D50=hybrid_func2_m_d50,
        )
        self.x_shift[9] = 0

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        D = x.shape[1]
        self.x_shift[0][1 : int(D / 2) : 2] = 5
        return super().__call__(x)


# CEC05 #21
class HybridCompositionFunction3(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                ExpandedScaffers_F6,
                ExpandedScaffers_F6,
                Rastrigin,
                Rastrigin,
                F8F2,
                F8F2,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
            ),
            sigmas=np.array([1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0], dtype=np.float64),
            lambdas=np.array(
                [5 * 5 / 100, 5 / 100, 5 * 1, 1, 5 * 1, 1, 5 * 10, 10, 5 * 5 / 200, 5 / 200],
                dtype=np.float64,
            ),
            fbias=fbias_data[20],
            x_shift=hybrid_func3_data,
            M_D2=hybrid_func3_m_d2,
            M_D10=hybrid_func3_m_d10,
            M_D30=hybrid_func3_m_d30,
            M_D50=hybrid_func3_m_d50,
        )


# CEC05 #22
class HybridCompositionFunction3H(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                ExpandedScaffers_F6,
                ExpandedScaffers_F6,
                Rastrigin,
                Rastrigin,
                F8F2,
                F8F2,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
            ),
            sigmas=np.array([1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0], dtype=np.float64),
            lambdas=np.array(
                [5 * 5 / 100, 5 / 100, 5 * 1, 1, 5 * 1, 1, 5 * 10, 10, 5 * 5 / 200, 5 / 200],
                dtype=np.float64,
            ),
            fbias=fbias_data[21],
            x_shift=hybrid_func3_data,
            M_D2=hybrid_func3_hm_d2,
            M_D10=hybrid_func3_hm_d10,
            M_D30=hybrid_func3_hm_d30,
            M_D50=hybrid_func3_hm_d50,
        )


# CEC05 #23
class NonContinuousHybridCompositionFunction3(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                ExpandedScaffers_F6,
                ExpandedScaffers_F6,
                Rastrigin,
                Rastrigin,
                F8F2,
                F8F2,
                Weierstrass,
                Weierstrass,
                Griewank,
                Griewank,
            ),
            sigmas=np.array([1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0], dtype=np.float64),
            lambdas=np.array(
                [5 * 5 / 100, 5 / 100, 5 * 1, 1, 5 * 1, 1, 5 * 10, 10, 5 * 5 / 200, 5 / 200],
                dtype=np.float64,
            ),
            fbias=fbias_data[22],
            x_shift=hybrid_func3_data,
            M_D2=hybrid_func3_m_d2,
            M_D10=hybrid_func3_m_d10,
            M_D30=hybrid_func3_m_d30,
            M_D50=hybrid_func3_m_d50,
        )

    def __call__(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = x.shape
        axis = [1] * (len(shape) - 1) + [-1]
        o = self.x_shift[0][: shape[-1]].reshape(axis)
        cond = np.abs(x - o) >= 0.5
        x[cond] = (np.round(2 * x) / 2)[cond]
        return super().__call__(x)


# CEC05 #24
class HybridCompositionFunction4(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Weierstrass,
                ExpandedScaffers_F6,
                F8F2,
                Ackley,
                Rastrigin,
                Griewank,
                NonContinuosExpandedScaffers_F6,
                NonContinuosRastrigin,
                HighConditionedElliptic,
                SphereWithNoise,
            ),
            sigmas=np.array([2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0], dtype=np.float64),
            lambdas=np.array(
                [10, 5 / 20, 1, 5 / 32, 1, 5 / 100, 5 / 50, 1, 5 / 100, 5 / 100], dtype=np.float64
            ),
            fbias=fbias_data[23],
            x_shift=hybrid_func4_data,
            M_D2=hybrid_func4_m_d2,
            M_D10=hybrid_func4_m_d10,
            M_D30=hybrid_func4_m_d30,
            M_D50=hybrid_func4_m_d50,
        )


# CEC05 #25
class HybridCompositionFunction4withoutbounds(SampleHybridCompositionFunction):
    def __init__(self) -> None:
        SampleHybridCompositionFunction.__init__(
            self,
            basic_functions=(
                Weierstrass,
                ExpandedScaffers_F6,
                F8F2,
                Ackley,
                Rastrigin,
                Griewank,
                NonContinuosExpandedScaffers_F6,
                NonContinuosRastrigin,
                HighConditionedElliptic,
                SphereWithNoise,
            ),
            sigmas=np.array([2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0], dtype=np.float64),
            lambdas=np.array(
                [10, 5 / 20, 1, 5 / 32, 1, 5 / 100, 5 / 50, 1, 5 / 100, 5 / 100], dtype=np.float64
            ),
            fbias=fbias_data[24],
            x_shift=hybrid_func4_data,
            M_D2=hybrid_func4_m_d2,
            M_D10=hybrid_func4_m_d10,
            M_D30=hybrid_func4_m_d30,
            M_D50=hybrid_func4_m_d50,
        )



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base import UniversalSet
from ..base._net import Net
from ..classifiers import GeneticProgrammingNeuralNetClassifier
from ..optimizers import GeneticProgramming
from ..optimizers import SHADE
from ..optimizers import SelfCGP
from ..tools.metrics import root_mean_square_error2d
from ..tools.random import train_test_split
from ..classifiers._gpnnclassifier import weights_type_optimizer_alias


class GeneticProgrammingNeuralNetRegressor(GeneticProgrammingNeuralNetClassifier):
    def __init__(
        self,
        iters: int,
        pop_size: int,
        input_block_size: int = 1,
        max_hidden_block_size: int = 9,
        offset: bool = True,
        output_activation: str = "sigma",
        test_sample_ratio: float = 0.5,
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
        weights_optimizer: weights_type_optimizer_alias = SHADE,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
        cache: bool = True,
    ):
        GeneticProgrammingNeuralNetClassifier.__init__(
            self,
            iters=iters,
            pop_size=pop_size,
            input_block_size=input_block_size,
            max_hidden_block_size=max_hidden_block_size,
            offset=offset,
            output_activation=output_activation,
            test_sample_ratio=test_sample_ratio,
            optimizer=optimizer,
            optimizer_args=optimizer_args,
            weights_optimizer=weights_optimizer,
            weights_optimizer_args=weights_optimizer_args,
            cache=cache,
        )

    def _evaluate_nets(
        self: GeneticProgrammingNeuralNetRegressor,
        weights: NDArray[np.float64],
        net: Net,
        X: NDArray[np.float64],
        targets: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        output2d = net.forward(X, weights)[:, :, 0]
        error = root_mean_square_error2d(targets, output2d)
        return error

    def _fitness_function(
        self: GeneticProgrammingNeuralNetRegressor,
        population: NDArray,
        X: NDArray[np.float64],
        targets: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        output2d = np.array([net.forward(X)[0] for net in population], dtype=np.float64)[:, :, 0]
        fitness = root_mean_square_error2d(targets, output2d)
        return fitness

    def _fit(
        self: GeneticProgrammingNeuralNetClassifier,
        X: NDArray[np.float64],
        y: NDArray[Union[np.float64, np.int64]],
    ) -> GeneticProgrammingNeuralNetClassifier:
        optimizer_args: dict[str, Any]

        if self._offset:
            X = np.hstack([X.copy(), np.ones((X.shape[0], 1))])

        n_outputs: int = 1

        X_train, X_test, y_train, y_test = train_test_split(X, y, self._test_sample_ratio)

        if self._optimizer_args is not None:
            assert (
                "iters" not in self._optimizer_args.keys()
                and "pop_size" not in self._optimizer_args.keys()
            ), """Do not set the "iters" or "pop_size" in the "optimizer_args". Instead,
              use the "SymbolicRegressionGP" arguments"""
            for arg in (
                "fitness_function",
                "uniset",
                "minimization",
            ):
                assert (
                    arg not in self._optimizer_args.keys()
                ), f"""Do not set the "{arg}"
                to the "optimizer_args". It is defined automatically"""
            optimizer_args = self._optimizer_args.copy()

        else:
            optimizer_args = {}

        uniset: UniversalSet = self._get_uniset(X)

        optimizer_args["fitness_function"] = lambda population: self._fitness_function(
            population, X_test, y_test
        )
        optimizer_args["genotype_to_phenotype"] = lambda trees: self._genotype_to_phenotype(
            X_train, y_train, trees, n_outputs
        )
        print(y_train)
        optimizer_args["iters"] = self._iters
        optimizer_args["pop_size"] = self._pop_size
        optimizer_args["uniset"] = uniset
        optimizer_args["minimization"] = True

        self._optimizer = self._optimizer_class(**optimizer_args)
        self._optimizer.fit()

        return self

    def _prepare_output(
        self: GeneticProgrammingNeuralNetClassifier, output: NDArray[np.float64]
    ) -> Union[NDArray[np.float64], NDArray[np.int64]]:
        return output



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import numpy as np
from numpy.typing import NDArray

from ._differentialevolution import DifferentialEvolution
from ..tools import donothing
from ..tools import find_pbest_id
from ..tools.operators import binomial
from ..tools.operators import current_to_pbest_1_archive_p_min
from ..tools.random import randc01
from ..tools.random import randn01
from ..tools.transformations import bounds_control_mean
from ..tools.transformations import lehmer_mean


class SHADE(DifferentialEvolution):
    """Tanabe, Ryoji & Fukunaga, Alex. (2013). Success-history based parameter adaptation
    for Differential Evolution. 2013 IEEE Congress on Evolutionary Computation,
    CEC 2013. 71-78. 10.1109/CEC.2013.6557555."""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        left: NDArray[np.float64],
        right: NDArray[np.float64],
        elitism: bool = True,
        init_population: Optional[NDArray[np.float64]] = None,
        genotype_to_phenotype: Callable[[NDArray[np.float64]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            left=left,
            right=right,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        self._F: NDArray[np.float64]
        self._CR: NDArray[np.float64]
        self._population_g_archive_i: NDArray[np.float64] = np.zeros(shape=(0, len(self._left)))
        self._population_archive: NDArray[np.float64]
        self._pbest_id: NDArray[np.int64]
        self._H_size: int = pop_size
        self._H_F = np.full(self._H_size, 0.5, dtype=np.float64)
        self._H_CR = np.full(self._H_size, 0.5, dtype=np.float64)
        self._k: int = 0
        self._p: float = 0.05

    def _generate_F_CR(self: SHADE) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        F_i = np.zeros(self._pop_size, dtype=np.float64)
        CR_i = np.zeros(self._pop_size, dtype=np.float64)
        for i in range(self._pop_size):
            r_i = np.random.randint(0, self._H_size)
            u_F = self._H_F[r_i]
            u_CR = self._H_CR[r_i]
            F_i[i] = randc01(np.float64(u_F))
            CR_i[i] = randn01(np.float64(u_CR))
        return (F_i, CR_i)

    def _append_archive(
        self, archive: NDArray[np.float64], worse_g: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        archive = np.append(archive, worse_g, axis=0)
        if len(archive) > self._pop_size:
            np.random.shuffle(archive)
            archive = archive[: self._pop_size]
        return archive

    def _update_u_F(self, u_F: float, S_F: NDArray[np.float64]) -> float:
        if len(S_F):
            return lehmer_mean(S_F)
        return u_F

    def _update_u_CR(
        self, u_CR: float, S_CR: NDArray[np.float64], df: NDArray[np.float64]
    ) -> float:
        if len(S_CR):
            sum_ = np.sum(df)
            if sum_ > 0:
                weight_i = df / sum_
                return np.sum(weight_i * S_CR)
        return u_CR

    def _get_new_individ_g(
        self: SHADE,
        individ_g: NDArray[np.float64],
        F: float,
        CR: float,
    ) -> NDArray[np.float64]:
        mutant_g = current_to_pbest_1_archive_p_min(
            individ_g, self._population_g_i, self._pbest_id, F, self._population_archive
        )

        mutant_cr_g = binomial(individ_g, mutant_g, np.float64(CR))
        return bounds_control_mean(mutant_cr_g, self._left, self._right)

    def _get_new_population(self: SHADE) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
        )

        self._F, self._CR = self._generate_F_CR()
        self._pbest_id = find_pbest_id(self._fitness_i, np.float64(self._p))
        self._population_archive = np.vstack([self._population_g_i, self._population_g_archive_i])

        mutant_cr_b_g = np.array(
            [
                get_new_individ_g(individ_g=self._population_g_i[i], F=self._F[i], CR=self._CR[i])
                for i in range(self._pop_size)
            ],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i
        succeses = mutant_cr_fit > self._fitness_i

        succeses_F = self._F[succeses]
        succeses_CR = self._CR[succeses]

        will_be_replaced_pop = self._population_g_i[succeses].copy()
        will_be_replaced_fit = self._fitness_i[succeses].copy()

        self._population_g_archive_i = self._append_archive(
            self._population_g_archive_i, will_be_replaced_pop
        )

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]

        d_fitness = np.abs(will_be_replaced_fit - self._fitness_i[succeses])

        if self._k + 1 == self._H_size:
            next_k = 0
        else:
            next_k = self._k + 1

        self._H_F[next_k] = self._update_u_F(self._H_F[self._k], succeses_F)
        self._H_CR[next_k] = self._update_u_CR(self._H_CR[self._k], succeses_CR, d_fitness)

        if self._k == self._H_size - 1:
            self._k = 0
        else:
            self._k += 1

    def _update_data(self: SHADE) -> None:
        super()._update_data()
        self._update_stats(H_F=self._H_F, H_CR=self._H_CR)



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base import Tree
from ..base._ea import EvolutionaryAlgorithm
from ..tools import donothing
from ..tools.operators import empty_crossover
from ..tools.operators import flip_mutation
from ..tools.operators import growing_mutation
from ..tools.operators import one_point_crossover
from ..tools.operators import one_point_crossoverGP
from ..tools.operators import point_mutation
from ..tools.operators import proportional_selection
from ..tools.operators import rank_selection
from ..tools.operators import shrink_mutation
from ..tools.operators import standart_crossover
from ..tools.operators import swap_mutation
from ..tools.operators import tournament_selection
from ..tools.operators import two_point_crossover
from ..tools.operators import uniform_crossover
from ..tools.operators import uniform_crossoverGP
from ..tools.operators import uniform_prop_crossover
from ..tools.operators import uniform_prop_crossover_GP
from ..tools.operators import uniform_rank_crossover
from ..tools.operators import uniform_rank_crossover_GP
from ..tools.operators import uniform_tour_crossover
from ..tools.operators import uniform_tour_crossover_GP
from ..tools.random import binary_string_population
from ..tools.transformations import rank_data
from ..tools.transformations import scale_data


class GeneticAlgorithm(EvolutionaryAlgorithm):
    """Holland, J. H. (1992). Genetic algorithms. Scientific American, 267(1), 66-72"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        str_len: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selection: str = "tournament_5",
        crossover: str = "uniform_2",
        mutation: str = "weak",
        init_population: Optional[NDArray[np.byte]] = None,
        genotype_to_phenotype: Callable[[NDArray[np.byte]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            elitism=elitism,
            init_population=init_population,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )
        self._str_len: int = str_len
        self._tour_size: int = tour_size
        self._parents_num: int = parents_num
        self._mutation_rate: float = mutation_rate
        self._specified_selection: str = selection
        self._specified_crossover: str = crossover
        self._specified_mutation: str = mutation

        self._selection_pool: Dict[str, Tuple[Callable, Union[float, int]]] = {
            "proportional": (proportional_selection, 0),
            "rank": (rank_selection, 0),
            "tournament_k": (tournament_selection, self._tour_size),
            "tournament_3": (tournament_selection, 3),
            "tournament_5": (tournament_selection, 5),
            "tournament_7": (tournament_selection, 7),
        }

        self._crossover_pool: Dict[str, Tuple[Callable, Union[float, int]]] = {
            "empty": (empty_crossover, 1),
            "one_point": (one_point_crossover, 2),
            "two_point": (two_point_crossover, 2),
            "uniform_2": (uniform_crossover, 2),
            "uniform_7": (uniform_crossover, 7),
            "uniform_k": (uniform_crossover, self._parents_num),
            "uniform_prop_2": (uniform_prop_crossover, 2),
            "uniform_prop_7": (uniform_prop_crossover, 7),
            "uniform_prop_k": (uniform_prop_crossover, self._parents_num),
            "uniform_rank_2": (uniform_rank_crossover, 2),
            "uniform_rank_7": (uniform_rank_crossover, 7),
            "uniform_rank_k": (uniform_rank_crossover, self._parents_num),
            "uniform_tour_3": (uniform_tour_crossover, 3),
            "uniform_tour_7": (uniform_tour_crossover, 7),
            "uniform_tour_k": (uniform_tour_crossover, self._parents_num),
            "gp_standart": (standart_crossover, 2),
            "gp_one_point": (one_point_crossoverGP, 2),
            "gp_uniform_2": (uniform_crossoverGP, 2),
            "gp_uniform_7": (uniform_crossoverGP, 7),
            "gp_uniform_k": (uniform_crossoverGP, self._parents_num),
            "gp_uniform_prop_2": (uniform_prop_crossover_GP, 2),
            "gp_uniform_prop_7": (uniform_prop_crossover_GP, 7),
            "gp_uniform_prop_k": (uniform_prop_crossover_GP, self._parents_num),
            "gp_uniform_rank_2": (uniform_rank_crossover_GP, 2),
            "gp_uniform_rank_7": (uniform_rank_crossover_GP, 7),
            "gp_uniform_rank_k": (uniform_rank_crossover_GP, self._parents_num),
            "gp_uniform_tour_3": (uniform_tour_crossover_GP, 3),
            "gp_uniform_tour_7": (uniform_tour_crossover_GP, 7),
            "gp_uniform_tour_k": (uniform_tour_crossover_GP, self._parents_num),
        }

        self._mutation_pool: Dict[str, Tuple[Callable, Union[float, int], bool]] = {
            "weak": (flip_mutation, 1 / 3, False),
            "average": (flip_mutation, 1, False),
            "strong": (flip_mutation, 3, False),
            "custom_rate": (flip_mutation, self._mutation_rate, True),
            "gp_weak_point": (point_mutation, 0.25, False),
            "gp_average_point": (point_mutation, 1, False),
            "gp_strong_point": (point_mutation, 4, False),
            "gp_custom_rate_point": (point_mutation, self._mutation_rate, True),
            "gp_weak_grow": (growing_mutation, 0.25, False),
            "gp_average_grow": (growing_mutation, 1, False),
            "gp_strong_grow": (growing_mutation, 4, False),
            "gp_custom_rate_grow": (growing_mutation, self._mutation_rate, True),
            "gp_weak_swap": (swap_mutation, 0.25, False),
            "gp_average_swap": (swap_mutation, 1, False),
            "gp_strong_swap": (swap_mutation, 4, False),
            "gp_custom_rate_swap": (swap_mutation, self._mutation_rate, True),
            "gp_weak_shrink": (shrink_mutation, 0.25, False),
            "gp_average_shrink": (shrink_mutation, 1, False),
            "gp_strong_shrink": (shrink_mutation, 4, False),
            "gp_custom_rate_shrink": (shrink_mutation, self._mutation_rate, True),
        }

        self._fitness_scale_i: NDArray[np.float64]
        self._fitness_rank_i: NDArray[np.float64]

    def _get_init_population(self: GeneticAlgorithm) -> None:
        if self._init_population is None:
            self._population_g_i = binary_string_population(self._pop_size, self._str_len)
        else:
            self._population_g_i = self._init_population.copy()

    def _get_new_individ_g(
        self: GeneticAlgorithm,
        specified_selection: str,
        specified_crossover: str,
        specified_mutation: str,
    ) -> Union[Tree, np.byte]:
        selection_func, tour_size = self._selection_pool[specified_selection]
        crossover_func, quantity = self._crossover_pool[specified_crossover]
        mutation_func, proba, is_constant_rate = self._mutation_pool[specified_mutation]

        selected_id = selection_func(
            self._fitness_scale_i, self._fitness_rank_i, np.int64(tour_size), np.int64(quantity)
        )

        offspring_no_mutated = crossover_func(
            self._population_g_i[selected_id],
            self._fitness_scale_i[selected_id],
            self._fitness_rank_i[selected_id],
        )

        if is_constant_rate:
            proba = proba
        else:
            proba = proba / len(offspring_no_mutated)

        offspring = mutation_func(offspring_no_mutated, np.float64(proba))
        return offspring

    def _get_new_population(self: GeneticAlgorithm) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
            self._specified_selection,
            self._specified_crossover,
            self._specified_mutation,
        )

        self._population_g_i = np.array(
            [get_new_individ_g() for _ in range(self._pop_size)], dtype=self._population_g_i.dtype
        )

    def _from_population_g_to_fitness(self: GeneticAlgorithm) -> None:
        super()._from_population_g_to_fitness()

        self._fitness_scale_i = scale_data(self._fitness_i)
        self._fitness_rank_i = rank_data(self._fitness_i)

        self._adapt()



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base._net import Net
from ..classifiers import MLPEAClassifier
from ..optimizers import SHADE
from ..tools.metrics import root_mean_square_error2d
from ..classifiers._gpnnclassifier import weights_type_optimizer_alias


class MLPEARegressor(MLPEAClassifier):
    def __init__(
        self,
        iters: int,
        pop_size: int,
        hidden_layers: Tuple[int, ...],
        activation: str = "sigma",
        output_activation: str = "sigma",
        offset: bool = True,
        weights_optimizer: weights_type_optimizer_alias = SHADE,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
    ):
        MLPEAClassifier.__init__(
            self,
            iters=iters,
            pop_size=pop_size,
            hidden_layers=hidden_layers,
            activation=activation,
            output_activation=output_activation,
            offset=offset,
            weights_optimizer=weights_optimizer,
            weights_optimizer_args=weights_optimizer_args,
        )

    def _evaluate_nets(
        self: MLPEARegressor,
        weights: NDArray[np.float64],
        net: Net,
        X: NDArray[np.float64],
        targets: NDArray[Union[np.float64, np.int64]],
    ) -> NDArray[np.float64]:
        output2d = net.forward(X, weights)[:, :, 0]
        error = root_mean_square_error2d(targets, output2d)
        return error

    def _fitness_function(
        self: MLPEARegressor,
        population: NDArray,
        X: NDArray[np.float64],
        targets: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        output2d = np.array([net.forward(X)[0] for net in population], dtype=np.float64)[:, :, 0]
        fitness = root_mean_square_error2d(targets, output2d)
        return fitness

    def _fit(
        self: MLPEARegressor, X: NDArray[np.float64], y: NDArray[Union[np.float64, np.int64]]
    ) -> MLPEARegressor:
        if self._offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        n_inputs: int = X.shape[1]
        n_outputs: int = len(set(y))

        self._net = self._defitne_net(n_inputs, n_outputs)

        self._net._weights = self._train_net(self._net, X, y.astype(np.float64))
        return self

    def _predict(
        self: MLPEARegressor, X: NDArray[np.float64]
    ) -> NDArray[Union[np.float64, np.int64]]:
        if self._offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        output = self._net.forward(X)[0, :, 0]
        y_pred = output
        return y_pred



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..tools import donothing


class TheFittest:
    def __init__(self) -> None:
        self._genotype: Any
        self._phenotype: Any
        self._fitness: float = -np.inf
        self._no_update_counter: int = 0

    def _update(
        self,
        population_g: NDArray[Any],
        population_ph: NDArray[Any],
        fitness: NDArray[np.float64],
    ) -> None:
        temp_best_id = np.argmax(fitness)
        temp_best_fitness = fitness[temp_best_id]
        if temp_best_fitness > self._fitness:
            self._replace(
                new_genotype=population_g[temp_best_id],
                new_phenotype=population_ph[temp_best_id],
                new_fitness=temp_best_fitness,
            )
            self._no_update_counter = 0
        else:
            self._no_update_counter += 1

    def _replace(self, new_genotype: Any, new_phenotype: Any, new_fitness: float) -> None:
        self._genotype = new_genotype.copy()
        self._phenotype = new_phenotype.copy()
        self._fitness = new_fitness

    def get(self) -> Dict:
        return {
            "genotype": self._genotype.copy(),
            "phenotype": self._phenotype.copy(),
            "fitness": self._fitness,
        }


class Statistics(dict):
    def _update(self, arg: Dict[str, Any]) -> None:
        for key, value in arg.items():
            try:
                value_to_append = value.copy()
            except AttributeError:
                value_to_append = value
            if key not in self.keys():
                self[key] = [value_to_append]
            else:
                self[key].append(value_to_append)


class EvolutionaryAlgorithm:
    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        elitism: bool = True,
        init_population: Optional[
            Union[NDArray[Any], NDArray[np.byte], NDArray[np.float64]]
        ] = None,
        genotype_to_phenotype: Callable[[NDArray[Any]], NDArray[Any]] = donothing,
        optimal_value: Optional[Union[float, int]] = None,
        termination_error_value: Union[float, int] = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        self._fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]] = fitness_function
        self._iters: int = iters
        self._pop_size: int = pop_size
        self._elitism: bool = elitism
        self._init_population: Optional[
            Union[NDArray[Any], NDArray[np.byte], NDArray[np.float64]]
        ] = init_population
        self._genotype_to_phenotype: Callable = genotype_to_phenotype
        self._no_increase_num: Optional[int] = no_increase_num
        self._show_progress_each: Optional[int] = show_progress_each
        self._keep_history: bool = keep_history

        self._sign: int = -1 if minimization else 1
        self._aim: Union[float, int] = self._get_aim(optimal_value, termination_error_value)
        self._calls: int = 0

        self._thefittest: TheFittest = TheFittest()
        self._stats: Statistics = Statistics()

        self._population_g_i: Union[NDArray[Any], NDArray[np.byte], NDArray[np.float64]]
        self._population_ph_i: NDArray
        self._fitness_i: NDArray[np.float64]

    def _get_init_population(self: EvolutionaryAlgorithm) -> None:
        return None

    def _get_aim(
        self: EvolutionaryAlgorithm,
        optimal_value: Optional[Union[float, int]],
        termination_error_value: Union[float, int],
    ) -> Union[float, int]:
        if optimal_value is not None:
            return self._sign * optimal_value - termination_error_value
        else:
            return np.inf

    def _get_fitness(
        self: EvolutionaryAlgorithm, population_ph: NDArray[Any]
    ) -> NDArray[np.float64]:
        self._calls += len(population_ph)
        return self._sign * self._fitness_function(population_ph)

    def _show_progress(self: EvolutionaryAlgorithm, current_iter: int) -> None:
        if self._show_progress_each is not None:
            cond_show_now = current_iter % self._show_progress_each == 0
            if cond_show_now:
                current_best = self._sign * self._thefittest._fitness
                print(f"{current_iter} iteration with fitness = {current_best}")

    def _termitation_check(self: EvolutionaryAlgorithm) -> bool:
        cond_aim = self._thefittest._fitness >= self._aim
        cond_no_increase = self._thefittest._no_update_counter == self._no_increase_num
        return bool(cond_aim or cond_no_increase)

    def _update_fittest(
        self: EvolutionaryAlgorithm,
        population_g: NDArray[Any],
        population_ph: NDArray[Any],
        fitness: NDArray[np.float64],
    ) -> None:
        self._thefittest._update(
            population_g=population_g, population_ph=population_ph, fitness=fitness
        )

    def _update_stats(self: EvolutionaryAlgorithm, **kwargs: Any) -> None:
        if self._keep_history:
            self._stats._update(kwargs)

    def _get_phenotype(self, popultion_g: NDArray[Any]) -> NDArray[Any]:
        return self._genotype_to_phenotype(popultion_g)

    def get_remains_calls(self: EvolutionaryAlgorithm) -> int:
        return (self._pop_size * self._iters) - self._calls

    def get_fittest(self: EvolutionaryAlgorithm) -> Dict:
        return self._thefittest.get()

    def get_stats(self: EvolutionaryAlgorithm) -> Statistics:
        return self._stats

    def _update_data(self: EvolutionaryAlgorithm) -> None:
        self._update_fittest(self._population_g_i, self._population_ph_i, self._fitness_i)
        self._update_stats(population_g=self._population_g_i, fitness_max=self._thefittest._fitness)

    def _adapt(self: EvolutionaryAlgorithm) -> None:
        return None

    def _from_population_g_to_fitness(self: EvolutionaryAlgorithm) -> None:
        self._population_ph_i = self._get_phenotype(self._population_g_i)
        self._fitness_i = self._get_fitness(self._population_ph_i)

        self._update_data()

        if self._elitism:
            (
                self._population_g_i[-1],
                self._population_ph_i[-1],
                self._fitness_i[-1],
            ) = self._thefittest.get().values()

    def _get_new_population(self: EvolutionaryAlgorithm) -> None:
        return None

    def fit(self: EvolutionaryAlgorithm) -> EvolutionaryAlgorithm:
        self._get_init_population()
        self._from_population_g_to_fitness()

        for i in range(self._iters - 1):
            self._show_progress(i)
            if self._termitation_check():
                break
            else:
                self._get_new_population()
                self._from_population_g_to_fitness()

        return self



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import numpy as np
from numpy.typing import NDArray

from ..base._ea import EvolutionaryAlgorithm
from ..tools import donothing
from ..tools.operators import binomialGA
from ..tools.operators import flip_mutation
from ..tools.operators import tournament_selection
from ..tools.random import binary_string_population
from ..tools.random import cauchy_distribution
from ..tools.transformations import lehmer_mean


class SHAGA(EvolutionaryAlgorithm):
    """Stanovov, Vladimir & Akhmedova, Shakhnaz & Semenkin, Eugene. (2019).
    Genetic Algorithm with Success History based Parameter Adaptation. 180-187.
    10.5220/0008071201800187."""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        str_len: int,
        elitism: bool = True,
        init_population: Optional[NDArray[np.byte]] = None,
        genotype_to_phenotype: Callable[[NDArray[np.byte]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        self._str_len = str_len
        self._MR: NDArray[np.float64]
        self._CR: NDArray[np.float64]
        self._H_size: int = pop_size
        self._H_MR = np.full(self._H_size, 1 / (self._str_len), dtype=np.float64)
        self._H_CR = np.full(self._H_size, 0.5, dtype=np.float64)
        self._k: int = 0

    def _get_init_population(self: SHAGA) -> None:
        if self._init_population is None:
            self._population_g_i = binary_string_population(self._pop_size, self._str_len)
        else:
            self._population_g_i = self._init_population.copy()
        self._population_ph_i = self._get_phenotype(self._population_g_i)
        self._fitness_i = self._get_fitness(self._population_ph_i)

    def _randc(self: SHAGA, u: float, scale: float) -> NDArray[np.float64]:
        value = cauchy_distribution(loc=u, scale=scale, size=1)[0]
        while value <= 0 or value > 5 / self._str_len:
            value = cauchy_distribution(loc=u, scale=scale, size=1)[0]
        return value

    def _randn(self: SHAGA, u: float, scale: float) -> float:
        value = np.random.normal(u, scale)
        if value < 0:
            value = 0
        elif value > 1:
            value = 1
        return value

    def _generate_MR_CR(self: SHAGA) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        MR_i = np.zeros(self._pop_size)
        CR_i = np.zeros(self._pop_size)
        for i in range(self._pop_size):
            r_i = np.random.randint(0, self._H_size)
            u_MR = self._H_MR[r_i]
            u_CR = self._H_CR[r_i]
            MR_i[i] = self._randc(u_MR, 0.1 / self._str_len)
            CR_i[i] = self._randn(u_CR, 0.1)
        return MR_i, CR_i

    def _update_u(self: SHAGA, u: float, S: NDArray[np.float64], df: NDArray[np.float64]) -> float:
        if len(S):
            sum_ = np.sum(df)
            if sum_ > 0:
                weight_i = df / sum_
                return lehmer_mean(x=S, weight=weight_i)
        return u

    def _get_new_individ_g(
        self: SHAGA,
        individ_g: NDArray[np.float64],
        MR: float,
        CR: float,
    ) -> NDArray[np.float64]:
        second_parent_id = tournament_selection(self._fitness_i, self._fitness_i, 2, 1)[0]
        second_parent = self._population_g_i[second_parent_id].copy()
        offspring = binomialGA(individ_g, second_parent, CR)
        mutant = flip_mutation(offspring, MR)
        return mutant

    def _get_new_population(self: SHAGA) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
        )
        self._MR, self._CR = self._generate_MR_CR()

        mutant_cr_b_g = np.array(
            [
                get_new_individ_g(individ_g=self._population_g_i[i], MR=self._MR[i], CR=self._CR[i])
                for i in range(self._pop_size)
            ],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i
        succeses = mutant_cr_fit > self._fitness_i

        succeses_MR = self._MR[succeses]
        succeses_CR = self._CR[succeses]

        will_be_replaced_fit = self._fitness_i[succeses].copy()

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]

        d_fitness = np.abs(will_be_replaced_fit - self._fitness_i[succeses])

        if self._k + 1 == self._H_size:
            next_k = 0
        else:
            next_k = self._k + 1

        self._H_MR[next_k] = self._update_u(self._H_MR[self._k], succeses_MR, d_fitness)
        self._H_CR[next_k] = self._update_u(self._H_CR[self._k], succeses_CR, d_fitness)

        if self._k == self._H_size - 1:
            self._k = 0
        else:
            self._k += 1

    def _update_data(self: SHAGA) -> None:
        super()._update_data()
        self._update_stats(H_MR=self._H_MR, H_CR=self._H_CR)

    def _from_population_g_to_fitness(self: EvolutionaryAlgorithm) -> None:
        self._update_data()

        if self._elitism:
            (
                self._population_g_i[-1],
                self._population_ph_i[-1],
                self._fitness_i[-1],
            ) = self._thefittest.get().values()

        self._adapt()



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base._model import Model
from ..base._net import ACTIV_NAME_INV
from ..base._net import Net
from ..classifiers._gpnnclassifier import weights_optimizer_alias
from ..classifiers._gpnnclassifier import weights_type_optimizer_alias
from ..optimizers import DifferentialEvolution
from ..optimizers import SHADE
from ..optimizers import jDE
from ..tools.metrics import categorical_crossentropy3d
from ..tools.random import float_population
from ..tools.transformations import GrayCode


class MLPEAClassifier(Model):
    def __init__(
        self,
        iters: int,
        pop_size: int,
        hidden_layers: Tuple[int, ...],
        activation: str = "sigma",
        output_activation: str = "softmax",
        offset: bool = True,
        weights_optimizer: weights_type_optimizer_alias = SHADE,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
    ):
        Model.__init__(self)

        self._iters: int = iters
        self._pop_size: int = pop_size
        self._hidden_layers: Tuple[int, ...] = hidden_layers
        self._activation: str = activation
        self._output_activation: str = output_activation
        self._offset: bool = offset
        self._weights_optimizer: weights_optimizer_alias
        self._weights_optimizer_class: weights_type_optimizer_alias = weights_optimizer
        self._weights_optimizer_args: Optional[dict[str, Any]] = weights_optimizer_args

    def _defitne_net(self: MLPEAClassifier, n_inputs: int, n_outputs: int) -> Net:
        start = 0
        end = n_inputs
        inputs_id = set(range(start, end))

        net = Net(inputs=inputs_id)

        for n_layer in self._hidden_layers:
            start = end
            end = end + n_layer
            inputs_id = {(n_inputs - 1)}
            hidden_id = set(range(start, end))
            activs = dict(zip(hidden_id, [ACTIV_NAME_INV[self._activation]] * len(hidden_id)))

            if self._offset:
                layer_net = Net(inputs=inputs_id) > Net(hidden_layers=[hidden_id], activs=activs)
            else:
                layer_net = Net(hidden_layers=[hidden_id], activs=activs)

            net = net > layer_net

        start = end
        end = end + n_outputs
        inputs_id = {(n_inputs - 1)}
        output_id = set(range(start, end))
        activs = dict(zip(output_id, [ACTIV_NAME_INV[self._output_activation]] * len(output_id)))

        if self._offset:
            layer_net = Net(inputs=inputs_id) > Net(outputs=output_id, activs=activs)
        else:
            layer_net = Net(outputs=output_id, activs=activs)

        net = net > layer_net
        net._get_order()
        return net

    def _evaluate_nets(
        self: MLPEAClassifier,
        weights: NDArray[np.float64],
        net: Net,
        X: NDArray[np.float64],
        targets: NDArray[Union[np.float64, np.int64]],
    ) -> NDArray[np.float64]:
        output3d = net.forward(X, weights)
        error = categorical_crossentropy3d(targets, output3d)
        return error

    def _train_net(
        self: MLPEAClassifier,
        net: Net,
        X_train: NDArray[np.float64],
        proba_train: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if self._weights_optimizer_args is not None:
            for arg in (
                "fitness_function",
                "left",
                "right",
                "str_len",
                "genotype_to_phenotype",
                "minimization",
            ):
                assert (
                    "iters" not in self._weights_optimizer_args.keys()
                    and "pop_size" not in self._weights_optimizer_args.keys()
                ), """Do not set the "iters" or "pop_size", or "uniset" in the "optimizer_args". Instead, use the "MLPClassifierEA" arguments"""
                assert (
                    arg not in self._weights_optimizer_args.keys()
                ), f"""Do not set the "{arg}"
              to the "weights_optimizer_args". It is defined automatically"""
            weights_optimizer_args = self._weights_optimizer_args.copy()
        else:
            weights_optimizer_args = {}

        weights_optimizer_args["iters"] = self._iters
        weights_optimizer_args["pop_size"] = self._pop_size
        left: NDArray[np.float64] = np.full(
            shape=len(net._weights), fill_value=-10, dtype=np.float64
        )
        right: NDArray[np.float64] = np.full(
            shape=len(net._weights), fill_value=10, dtype=np.float64
        )
        initial_population: Union[NDArray[np.float64], NDArray[np.byte]] = float_population(
            weights_optimizer_args["pop_size"], left, right
        )
        initial_population[0] = net._weights.copy()
        weights_optimizer_args["fitness_function"] = lambda population: self._evaluate_nets(
            population, net, X_train, proba_train
        )
        if self._weights_optimizer_class in (SHADE, DifferentialEvolution, jDE):
            weights_optimizer_args["left"] = left
            weights_optimizer_args["right"] = right
        else:
            parts: NDArray[np.int64] = np.full(
                shape=len(net._weights), fill_value=16, dtype=np.int64
            )
            genotype_to_phenotype = GrayCode(fit_by="parts").fit(left, right, parts)
            weights_optimizer_args["str_len"] = np.sum(parts)
            weights_optimizer_args["genotype_to_phenotype"] = genotype_to_phenotype.transform

        weights_optimizer_args["minimization"] = True
        optimizer = self._weights_optimizer_class(**weights_optimizer_args)
        optimizer.fit()

        phenotype = optimizer.get_fittest()["phenotype"]

        self._weights_optimizer = optimizer

        return phenotype

    def get_optimizers(self: MLPEAClassifier) -> Tuple:
        return (self._weights_optimizer,)

    def _fit(
        self: MLPEAClassifier, X: NDArray[np.float64], y: NDArray[Union[np.float64, np.int64]]
    ) -> MLPEAClassifier:
        if self._offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        n_inputs: int = X.shape[1]
        n_outputs: int = len(set(y))
        eye: NDArray[np.float64] = np.eye(n_outputs, dtype=np.float64)
        target_probas: NDArray[np.float64] = eye[y]

        self._net = self._defitne_net(n_inputs, n_outputs)

        self._net._weights = self._train_net(self._net, X, target_probas)
        return self

    def _predict(
        self: MLPEAClassifier, X: NDArray[np.float64]
    ) -> NDArray[Union[np.float64, np.int64]]:
        if self._offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        output = self._net.forward(X)[0]
        y_pred = np.argmax(output, axis=1)
        return y_pred



from ._tree import EphemeralNode
from ._tree import FunctionalNode
from ._tree import TerminalNode
from ._tree import Tree
from ._tree import UniversalSet


__all__ = ["Tree", "FunctionalNode", "TerminalNode", "EphemeralNode", "UniversalSet"]



import numpy as np

from ..base._ea import Statistics
from ..base._ea import TheFittest
from ..base._net import Net
from ..base._net import HiddenBlock
from ..base._tree import EphemeralConstantNode
from ..base._tree import EphemeralNode
from ..base._tree import FunctionalNode
from ..base._tree import TerminalNode
from ..base._tree import Tree
from ..base._tree import UniversalSet
from ..tools.operators import Add
from ..tools.operators import Mul
from ..tools.operators import Operator
from ..tools.operators import Sub


class PlusOne(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({} + 1)", name="plus_one", sign="+1")

    def __call__(self, x: int) -> int:
        result = x + 1
        return result


def test_thefittest_class():
    net_1 = Net()
    net_2 = Net()
    net_3 = Net()

    population_g = np.array([net_1, net_2, net_3], dtype=object)
    population_ph = population_g
    fitness = np.array([1, 2, 3], dtype=np.float64)

    thefittest_ = TheFittest()
    thefittest_._update(population_g, population_ph, fitness)
    assert thefittest_._fitness == 3
    assert thefittest_._no_update_counter == 0
    assert thefittest_._fitness is not fitness[2]
    assert thefittest_._genotype is not population_g[2]
    assert thefittest_._phenotype is not population_ph[2]

    fitness = np.array([1, 2, 2], dtype=np.float64)
    thefittest_._update(population_g, population_ph, fitness)
    assert thefittest_._no_update_counter == 1
    assert thefittest_._fitness == 3

    fitness = np.array([1, 2, 4], dtype=np.float64)
    thefittest_._update(population_g, population_ph, fitness)
    assert thefittest_._no_update_counter == 0
    assert thefittest_._fitness == 4

    fittest = thefittest_.get()
    assert isinstance(fittest, dict)
    assert len(fittest) == 3


def test_statistic_class():
    net_1 = Net()
    net_2 = Net()
    net_3 = Net()
    population_g = np.array([net_1, net_2, net_3], dtype=object)
    population_ph = population_g
    fitness = np.array([1, 2, 3], dtype=np.float64)

    statistics_ = Statistics()
    fitness_max = np.max(fitness)
    statistics_._update(
        {"population_g": population_g, "population_ph": population_ph, "fitness_max": fitness_max}
    )

    assert statistics_["fitness_max"][0] == fitness_max
    assert np.all(statistics_["population_g"][0] == population_g)
    assert np.all(statistics_["population_ph"][0] == population_ph)

    statistics_._update(
        {"population_g": population_g, "population_ph": population_ph, "fitness_max": fitness_max}
    )

    assert len(statistics_["fitness_max"]) == 2
    assert len(statistics_["population_g"]) == 2
    assert len(statistics_["population_ph"]) == 2


def test_nodes_class():
    def generator():
        return 10

    functional_node = FunctionalNode(value=Add(), sign="+")
    assert functional_node._n_args == 2
    assert functional_node._sign == "+"
    assert functional_node._value(1, 2) == 3

    functional_node2 = FunctionalNode(value=PlusOne())
    assert functional_node2._sign == "+1"
    assert functional_node2._value(1) == 2

    functional_node3 = FunctionalNode(value=Add())
    assert functional_node3._sign == Add()._sign

    terminal_node = TerminalNode(value=1, name="1")
    assert str(terminal_node) == "1"

    terminal_node2 = TerminalNode(value=2, name="2")

    assert terminal_node != terminal_node2

    ephemeral_node = EphemeralNode(generator)
    ephemeral_constant_node = ephemeral_node()
    assert isinstance(ephemeral_constant_node, EphemeralConstantNode)

    universal_set = UniversalSet(
        functional_set=(functional_node, functional_node2, functional_node3),
        terminal_set=(terminal_node, terminal_node2),
    )

    random_terminal = universal_set._random_terminal_or_ephemeral()
    assert random_terminal in universal_set._terminal_set

    random_functional_2 = universal_set._random_functional(n_args=2)
    assert random_functional_2 in universal_set._functional_set[2]
    assert random_functional_2._n_args == 2

    random_functional_1 = universal_set._random_functional(n_args=1)
    assert random_functional_1 in universal_set._functional_set[1]
    assert random_functional_1._n_args == 1
    assert random_functional_1 is functional_node2

    universal_set = UniversalSet(
        functional_set=(functional_node, functional_node2, functional_node3),
        terminal_set=(ephemeral_node,),
    )
    random_ephemeral = universal_set._random_terminal_or_ephemeral()
    assert isinstance(random_ephemeral, EphemeralConstantNode)
    assert random_ephemeral._value == 10


def test_tree_class():
    x = TerminalNode(value=5, name="x")
    y = TerminalNode(value=7, name="y")
    z = TerminalNode(value=11, name="z")

    functional_add = FunctionalNode(value=Add(), sign="+")
    functional_mul = FunctionalNode(value=Mul(), sign="*")
    functional_sub = FunctionalNode(value=Sub(), sign="-")

    tree = Tree(nodes=[functional_mul, z, functional_add, x, y])

    tree2 = Tree(
        nodes=[functional_add, functional_mul, x, x, functional_mul, z, functional_sub, y, x]
    )

    assert tree() == 132
    assert str(tree) == "(z * (x + y))"
    assert len(tree) == 5

    assert tree2() == 47
    assert str(tree2) == "((x * x) + (z * (y - x)))"
    assert len(tree2) == 9

    assert tree2 == tree2
    assert bool(tree2 == tree) is False

    tree_copy = tree.copy()
    assert tree_copy == tree
    assert tree_copy is not tree
    assert tree_copy._nodes is not tree._nodes
    assert tree_copy._n_args is not tree._n_args

    tree3 = tree2.set_terminals(x=3, y=4, z=5)
    assert tree3() == 14
    assert bool(tree2 == tree3) is True

    tree3._nodes[0] = functional_mul
    assert bool(tree2 == tree3) is False

    index, n_index = tree2.subtree_id(4)
    assert (index, n_index) == (4, 9)

    tree4 = tree2.subtree(4)
    assert tree4() == 22
    assert str(tree4) == "(z * (y - x))"
    assert len(tree) == 5

    tree5 = tree.concat(index=4, other_tree=tree)
    assert str(tree5) == "(z * (x + (z * (x + y))))"
    assert tree5() == 1507

    args_id = tree.get_args_id(index=2)
    assert np.all(args_id == np.array([3, 4], dtype=np.int64))

    levels = tree.get_levels(index=0)
    assert np.all(levels == np.array([0, 1, 1, 2, 2], dtype=np.int64))

    max_level = tree.get_max_level()
    assert max_level == 2

    graph = tree.get_graph()
    graph = tree.get_graph(keep_id=True)

    assert isinstance(graph, dict)


def test_net():
    net = Net(inputs={0, 1, 3})
    net2 = Net(inputs={2})

    net3 = net + net2
    assert net3._inputs == {0, 1, 2, 3}

    net4 = net > net2
    assert net4._inputs == {0, 1, 2, 3}

    net5 = Net(hidden_layers=[{4}])
    net6 = Net(hidden_layers=[{5, 6}])

    net7 = net5 + net6

    assert net7._hidden_layers == [{4, 5, 6}]

    net8 = net5 > net6
    assert net8._hidden_layers == [{4}, {5, 6}]
    assert np.all(net8._connects == np.array([[4, 5], [4, 6]], dtype=np.int64))

    net9 = net2 + net5
    assert np.all(net9._connects == np.array([[2, 4]], dtype=np.int64))

    net10 = net5 + net2
    assert np.all(net10._connects == np.array([[2, 4]], dtype=np.int64))

    net11 = net5 > net2
    assert np.all(net11._connects == np.array([[2, 4]], dtype=np.int64))

    net12 = Net(hidden_layers=[{4}, {5, 6}])
    net13 = Net(hidden_layers=[{7, 8}])

    net14 = net12 + net13
    assert net14._hidden_layers == [{4, 7, 8}, {5, 6}]

    net15 = net13 + net12
    assert net15._hidden_layers == [{4, 7, 8}, {5, 6}]

    connects, weights = net._get_connect({1, 2}, {})
    assert np.all(connects == np.zeros((0, 2), dtype=int))
    assert np.all(weights == np.zeros((0), dtype=float))

    assert len(net11) == 1

    fixed = net5._fix(inputs={1, 2})

    assert np.all(fixed._connects == np.array([[1, 4], [2, 4]], dtype=np.int64))
    assert fixed._inputs == {1, 2}

    net16 = Net(
        inputs={0, 1},
        hidden_layers=[{2, 3}, {4}],
        outputs={5},
        connects=np.array([[0, 2], [1, 3], [2, 4], [3, 4], [4, 5]], dtype=np.int64),
        weights=np.array([0.1, 0.2, 0.4, 0.5, 1.5], dtype=np.float64),
        activs={2: 1, 3: 1, 4: 1, 5: 1},
    )

    X = np.array([[1, 2]], dtype=np.float64)
    out = net16.forward(X)

    print(out)

    graph = net16.get_graph()

    assert out[0][0][0] == 0.36000000000000004
    assert isinstance(graph, dict)

    hb = HiddenBlock(max_size=10)

    assert hb._activ in [0, 1, 2, 3]
    assert hb._size <= 10



from typing import Union

from ._differentialevolution import DifferentialEvolution
from ._geneticalgorithm import GeneticAlgorithm
from ._geneticprogramming import GeneticProgramming
from ._jde import jDE
from ._selfcga import SelfCGA
from ._selfcgp import SelfCGP
from ._shade import SHADE
from ._shaga import SHAGA


_optimizer_tree_type = [GeneticProgramming, SelfCGP]

OptimizerTreeType = Union[GeneticProgramming, SelfCGP]


optimizer_real_coded = [DifferentialEvolution, SHADE, jDE]

OptimizerRealCoded = Union[DifferentialEvolution, SHADE, jDE]

optimizer_binary_coded = [GeneticAlgorithm, SelfCGA, SHAGA]

OptimizerBinaryCoded = Union[GeneticAlgorithm, SelfCGA, SHAGA]

optimizer_string_type = optimizer_real_coded + optimizer_binary_coded

OptimizerStringType = Union[OptimizerRealCoded, OptimizerBinaryCoded]

OptimizerAnyType = Union[OptimizerTreeType, OptimizerStringType]

optimizer_any_type = _optimizer_tree_type + optimizer_string_type

__all__ = [
    "DifferentialEvolution",
    "GeneticAlgorithm",
    "GeneticProgramming",
    "jDE",
    "SelfCGA",
    "SelfCGP",
    "SHADE",
    "SHAGA",
]
